# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
PhotonicCircuit class for creating circuits with Parameters object that can be
modified after creation.
"""

from collections.abc import Iterable
from copy import copy, deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Union

import matplotlib.pyplot as plt
import numpy as np
from IPython import display as ipy_display
from numpy.typing import NDArray

from lightworks.sdk.utils.exceptions import (
    CircuitCompilationError,
    LightworksDependencyError,
    ModeRangeError,
)
from lightworks.sdk.utils.param_unitary import ParameterizedUnitary
from lightworks.sdk.visualisation import display

from .parameters import Parameter
from .photonic_circuit_utils import (
    add_empty_mode_to_circuit_spec,
    add_modes_to_circuit_spec,
    check_loss,
    compress_mode_swaps,
    convert_non_adj_beamsplitters,
    find_optimal_mode_swapping,
    unpack_circuit_spec,
)
from .photonic_compiler import CompiledPhotonicCircuit
from .photonic_components import (
    Barrier,
    BeamSplitter,
    Component,
    Group,
    HeraldData,
    Loss,
    ModeSwaps,
    PhaseShifter,
)

if TYPE_CHECKING:
    from .unitary import Unitary


class PhotonicCircuit:
    """
    Provides support for building photonic circuits from a set of linear optic
    components, with the ability to assign certain quantities of components to
    Parameter objects whose values can be adjusted after creation.

    Args:

        n_modes (int) : The number of modes used in the circuit.

    """

    def __init__(self, n_modes: int) -> None:
        if not isinstance(n_modes, int):
            if int(n_modes) == n_modes:
                n_modes = int(n_modes)
            else:
                raise TypeError("Number of modes should be an integer.")
        self.__n_modes = n_modes
        self.__circuit_spec: list[Component] = []
        self.__in_heralds: dict[int, int] = {}
        self.__out_heralds: dict[int, int] = {}
        self.__external_in_heralds: dict[int, int] = {}
        self.__external_out_heralds: dict[int, int] = {}
        self.__internal_modes: list[int] = []

    def __add__(self, value: "PhotonicCircuit") -> "PhotonicCircuit":
        """Defines custom addition behaviour for two circuits."""
        if not isinstance(value, PhotonicCircuit):
            raise TypeError(
                "Addition only supported between two PhotonicCircuit objects."
            )
        if self.n_modes != value.n_modes:
            raise ModeRangeError(
                "Two circuits to add must have the same number of modes."
            )
        if self.heralds.input or value.heralds.input:
            raise NotImplementedError(
                "Support for heralds when combining circuits not yet "
                "implemented."
            )
        # Create new circuits and combine circuits specs to create new one
        new_circ = PhotonicCircuit(self.n_modes)
        new_circ.__circuit_spec = self.__circuit_spec + value.__circuit_spec
        return new_circ

    def __str__(self) -> str:
        return f"PhotonicCircuit({self.n_modes})"

    def __repr__(self) -> str:
        return f"lightworks.PhotonicCircuit({self.n_modes})"

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, PhotonicCircuit):
            return False
        if self.n_modes != value.n_modes:
            return False
        u1, u2 = self.U_full, value.U_full
        if u1.shape != u2.shape:
            return False
        return np.allclose(u1, u2, rtol=0, atol=1e-8)

    @property
    def U(self) -> NDArray[np.complex128]:  # noqa: N802
        """
        The effective unitary that the circuit implements across modes. This
        will include the effect of any loss within a circuit. It is calculated
        using the parameter values at the time that the attribute is called.
        """
        return self._build().U_full[: self.n_modes, : self.n_modes]

    @property
    def U_full(self) -> NDArray[np.complex128]:  # noqa: N802
        """
        The full unitary for the created circuit, this will include the
        additional modes used for the simulation of loss, if this has been
        included in the circuit.
        """
        return self._build().U_full

    @property
    def n_modes(self) -> int:
        """The number of modes in the circuit."""
        return self.__n_modes

    @n_modes.setter
    def n_modes(self, value: Any) -> None:  # noqa: ARG002
        """
        Prevents modification of n_modes attribute after circuit creation.
        """
        raise AttributeError("Number of circuit modes cannot be modified.")

    @property
    def input_modes(self) -> int:
        """
        The number of input modes that should be specified, accounting for the
        heralds used in the circuit.
        """
        return self.n_modes - len(self.heralds.input)

    @property
    def heralds(self) -> HeraldData:
        """
        A dictionary which details the set heralds on the inputs and outputs of
        the circuit.
        """
        return HeraldData(
            input=copy(self.__in_heralds), output=copy(self.__out_heralds)
        )

    @property
    def _internal_modes(self) -> list[int]:
        return self.__internal_modes

    @property
    def _external_heralds(self) -> HeraldData:
        """
        Stores details of heralds which are on the outside of a circuit (i.e.
        were not added as part of a group).
        """
        return HeraldData(
            input=copy(self.__external_in_heralds),
            output=copy(self.__external_out_heralds),
        )

    def add(
        self,
        circuit: Union["PhotonicCircuit", "Unitary"],
        mode: int = 0,
        group: bool = False,
        name: str | None = None,
    ) -> None:
        """
        Can be used to add either another PhotonicCircuit or a Unitary component
        to the existing circuit. This can either have the same size or be
        smaller than the circuit which is being added to.

        Args:

            circuit (PhotonicCircuit | Unitary) : The circuit/component that is
                to be added.

            mode (int, optional) : The first mode on which the circuit should
                be placed. If not specified it defaults to zero.

            group (bool, optional) : Used to control whether the circuit
                components are added to the existing circuit or placed within a
                group which contains all components in a single element.
                Defaults to False unless the added circuit has heralds, in
                which grouping is enforced.

            name (str | None, optional) : Set a name to use when displaying the
                added circuit. This is only applied when the group option is
                used.

        """
        if not isinstance(circuit, PhotonicCircuit):
            raise TypeError(
                "Add method only supported for PhotonicCircuit or Unitary "
                "objects."
            )
        # Remap mode
        mode = self._map_mode(mode)
        self._mode_in_range(mode)
        # Make copy of circuit to avoid modification
        circuit = circuit.copy()
        # Force grouping if heralding included
        group = True if circuit.heralds.input else group
        # When name not provided set this
        if name is None:
            spec = circuit.__circuit_spec
            # Check special case where name is retrieved from previous group
            name = (
                spec[0].name
                if len(spec) == 1 and isinstance(spec[0], Group)
                else "Circuit"
            )
        # When grouping use unpacked circuit
        if group:
            circuit.unpack_groups()
        # Check circuit size is valid
        n_heralds = len(circuit.heralds.input)
        if mode + circuit.n_modes - n_heralds > self.n_modes:
            raise ModeRangeError("Circuit to add is outside of mode range")

        # Include any existing internal modes into the circuit to be added
        for i in sorted(self.__internal_modes):
            # Need to account for shifts when adding new heralds
            target_mode = i - mode
            for m in circuit.heralds.input:
                if target_mode > m:
                    target_mode += 1
            if 0 <= target_mode < circuit.n_modes:
                circuit._add_empty_mode(target_mode)

        # Then add new modes for heralds from circuit and also add swaps to
        # enforce that the input and output herald are on the same mode
        provisional_swaps = {}
        circuit_heralds = circuit.heralds
        in_herald_modes = list(circuit_heralds.input.keys())
        out_herald_modes = list(circuit_heralds.output.keys())
        for m in sorted(in_herald_modes):
            self._add_empty_mode(mode + m)
            self.__internal_modes.append(mode + m)
            # Current limitation is that heralding should be on the same mode
            # when adding, so use a mode swap to compensate for this.
            herald_loc = in_herald_modes.index(m)
            out_herald = out_herald_modes[herald_loc]
            provisional_swaps[out_herald] = m

        # Convert provisional swaps into full list and add to circuit
        swaps = find_optimal_mode_swapping(provisional_swaps, circuit.n_modes)
        spec = circuit.__circuit_spec
        # Skip for cases where swaps do not alter mode structure
        if list(swaps.keys()) != list(swaps.values()):
            spec.append(ModeSwaps(swaps))
        # Update heralds to enforce input and output are on the same mode
        new_heralds = HeraldData(
            input=circuit.heralds.input,
            output={swaps[m]: n for m, n in circuit.heralds.output.items()},
        )
        # Also add all included heralds to the heralds dict
        for m in new_heralds.input:
            self.__in_heralds[m + mode] = new_heralds.input[m]
            self.__out_heralds[m + mode] = new_heralds.output[m]
        # And shift all components in circuit by required amount
        add_cs = add_modes_to_circuit_spec(spec, mode)

        # Then add circuit spec, adjusting how this is included
        if not group:
            self.__circuit_spec += add_cs
        else:
            self.__circuit_spec.append(
                Group(
                    add_cs, name, mode, mode + circuit.n_modes - 1, new_heralds
                )
            )

    def bs(
        self,
        mode_1: int,
        mode_2: int | None = None,
        reflectivity: float | Parameter[float] = 0.5,
        loss: float | Parameter[float] = 0,
        convention: str = "Rx",
    ) -> None:
        """
        Add a beam splitter of specified reflectivity between two modes to the
        circuit.

        Args:

            mode_1 (int) : The first mode which the beam splitter acts on.

            mode_2 (int | None, optional) : The second mode that the beam
                splitter acts on. This can also be left as the default value of
                None to automatically use mode_2 as mode_1 + 1.

            reflectivity (float | Parameter, optional) : The reflectivity value
                of the beam splitter. Defaults to 0.5.

            loss (float | Parameter, optional) : The loss of the beam splitter,
                this should be provided as a decimal value in the range [0,1].

            convention (str, optional) : The convention to use for the beam
                splitter, should be either "Rx" (the default value) or "H".

        """
        if mode_2 is None:
            mode_2 = mode_1 + 1
        mode_1 = self._map_mode(mode_1)
        self._mode_in_range(mode_1)
        mode_2 = self._map_mode(mode_2)
        if mode_1 == mode_2:
            raise ModeRangeError(
                "Beam splitter must act across two different modes."
            )
        self._mode_in_range(mode_2)
        # Validate loss before updating circuit spec
        check_loss(loss)
        # Then update circuit spec
        self.__circuit_spec.append(
            BeamSplitter(mode_1, mode_2, reflectivity, convention)
        )
        if isinstance(loss, Parameter) or loss > 0:
            self.loss(mode_1, loss)
            self.loss(mode_2, loss)

    def ps(
        self,
        mode: int,
        phi: float | Parameter[float],
        loss: float | Parameter[float] = 0,
    ) -> None:
        """
        Applies a phase shift to a given mode in the circuit.

        Args:

            mode (int) : The mode on which the phase shift acts.

            phi (float | Parameter) : The angular phase shift to apply.

            loss (float | Parameter, optional) : The loss of the phase shifter,
                this should be provided as a decimal value in the range [0,1].

        """
        mode = self._map_mode(mode)
        self._mode_in_range(mode)
        check_loss(loss)
        self.__circuit_spec.append(PhaseShifter(mode, phi))
        if isinstance(loss, Parameter) or loss > 0:
            self.loss(mode, loss)

    def loss(self, mode: int, loss: float | Parameter[float] = 0) -> None:
        """
        Adds a loss channel to the specified mode of the circuit.

        Args:

            mode (int) : The mode on which the loss channel acts.

            loss (float | Parameter, optional) : The loss applied to the
                selected mode, this should be provided as a decimal value in the
                range [0,1].

        """
        mode = self._map_mode(mode)
        self._mode_in_range(mode)
        check_loss(loss)
        self.__circuit_spec.append(Loss(mode, loss))

    def barrier(self, modes: list[int] | None = None) -> None:
        """
        Adds a barrier to separate different parts of a circuit. This is
        applied to the specified modes.

        Args:

            modes (list | None) : The modes over which the barrier should be
                applied to.

        """
        if modes is None:
            modes = list(range(self.n_modes - len(self.__internal_modes)))
        modes = [self._map_mode(i) for i in modes]
        for m in modes:
            self._mode_in_range(m)
        self.__circuit_spec.append(Barrier(modes))

    def mode_swaps(self, swaps: dict[int, int]) -> None:
        """
        Performs swaps between a given set of modes. The keys of the dictionary
        should correspond to the initial modes and the values to the modes they
        are swapped to. It is also required that all mode swaps are complete,
        i.e. any modes which are swapped to must also be sent to a target
        destination.

        Args:

            swaps (dict) : A dictionary detailing the original modes and the
                locations that they are to be swapped to.

        """
        # Remap swap dict and check modes
        swaps = {
            self._map_mode(mi): self._map_mode(mo) for mi, mo in swaps.items()
        }
        for m in [*swaps.keys(), *swaps.values()]:
            self._mode_in_range(m)
        self.__circuit_spec.append(ModeSwaps(swaps))

    def herald(
        self, modes: int | tuple[int, int], photons: int | tuple[int, int]
    ) -> None:
        """
        Add a herald across a selected input/output of the circuit. If only one
        mode is specified then this will be used for both the input and output.

        Args:

            modes (int | tuple) : The mode(s) to use for heralding. If supplied
                as a single value then the same mode will be used for both input
                and output, otherwise a pair of values can be specified to use
                a different input and output.

            photons (int | tuple) : The photon(s) to use for heralding. If
                supplied as a single value then the same number of photons will
                be heralded on the input and output, otherwise a pair of values
                can be specified to use a different number of photons.

        """
        if isinstance(photons, Iterable) and not isinstance(photons, str):
            in_photon, out_photon = photons
        else:
            in_photon = out_photon = photons
        for n in [in_photon, out_photon]:
            if not isinstance(n, int) or isinstance(n, bool):
                raise TypeError(
                    "Number of photons for herald should be a tuple of two "
                    "integers."
                )
        if isinstance(modes, Iterable) and not isinstance(photons, str):
            input_mode, output_mode = modes
        else:
            input_mode = output_mode = modes
        input_mode = self._map_mode(input_mode)
        output_mode = self._map_mode(output_mode)
        self._mode_in_range(input_mode)
        self._mode_in_range(output_mode)
        # Check if herald already used on input or output
        if input_mode in self.__in_heralds:
            raise ValueError("Heralding already set for chosen input mode.")
        if output_mode in self.__out_heralds:
            raise ValueError("Heralding already set for chosen output mode.")
        # If not then update dictionaries
        self.__in_heralds[input_mode] = in_photon
        self.__out_heralds[output_mode] = out_photon
        self.__external_in_heralds[input_mode] = in_photon
        self.__external_out_heralds[output_mode] = out_photon

    def display(
        self,
        show_parameter_values: bool = False,
        display_loss: bool = False,
        mode_labels: list[str] | None = None,
        display_type: Literal["svg", "mpl"] = "svg",
        display_barriers: bool = False,
    ) -> None:
        """
        Displays the current circuit with parameters set using either their
        current values or labels.

        Args:

            show_parameter_values (bool, optional) : Shows the values of
                parameters instead of the associated labels if specified.

            display_loss (bool, optional) : Choose whether to display loss
                components in the figure, defaults to False.

            mode_labels (list|None, optional) : Optionally provided a list of
                mode labels which will be used to name the mode something other
                than numerical values. Can be set to None to use default
                values.

            display_type (str, optional) : Selects whether the drawsvg or
                matplotlib module should be used for displaying the circuit.
                Should either be 'svg' or 'mpl', defaults to 'svg'.

            display_barriers (bool, optional) : Shows included barriers within
                the created visualization if this is required.

        """
        return_ = display(
            self,
            display_loss=display_loss,
            mode_labels=mode_labels,
            display_type=display_type,
            show_parameter_values=show_parameter_values,
            display_barriers=display_barriers,
        )
        if display_type == "mpl":
            plt.show()
        elif display_type == "svg":
            ipy_display.display(return_)

    def save_figure(
        self,
        path: str | Path,
        svg: bool = True,
        show_parameter_values: bool = False,
        display_loss: bool = False,
        mode_labels: list[str] | None = None,
        display_barriers: bool = False,
    ) -> None:
        """
        Creates a figure of the current circuit and saves this to the provided
        file name.

        Args:

            path (str|Path) : The path to save the figure to. This can be just a
                filename or can also include a directory which the circuit
                should be placed within.

            svg (bool, optional) : Controls whether the figure is saved as an
                svg (True) or png (False). Defaults to svg.

            show_parameter_values (bool, optional) : Shows the values of
                parameters instead of the associated labels if specified.

            display_loss (bool, optional) : Choose whether to display loss
                components in the figure, defaults to False.

            mode_labels (list|None, optional) : Optionally provided a list of
                mode labels which will be used to name the mode something other
                than numerical values. Can be set to None to use default
                values.

            display_barriers (bool, optional) : Shows included barriers within
                the created visualization if this is required.

        """
        figure = display(
            self,
            display_loss=display_loss,
            mode_labels=mode_labels,
            display_type="svg",
            show_parameter_values=show_parameter_values,
            display_barriers=display_barriers,
        )
        p = Path(path)
        p = p.with_suffix(".svg") if svg else p.with_suffix(".png")
        p.parent.mkdir(parents=True, exist_ok=True)
        if svg:
            figure.save_svg(p)
        else:
            figure.set_pixel_scale(5)
            try:
                figure.save_png(str(p))
            except ImportError as e:
                raise LightworksDependencyError(
                    "Unable to save figure as png due to missing dependency."
                    "Please see original exception for more information or "
                    "seek support. Alternatively, it will still be possible to "
                    "save the figure as an svg."
                ) from e

    def get_all_params(self) -> list[Parameter[Any]]:
        """
        Returns all the Parameter objects used as part of creating the circuit.
        """
        all_params = []
        for spec in unpack_circuit_spec(self.__circuit_spec):
            for p in spec.values():
                if isinstance(p, Parameter) and p not in all_params:
                    all_params.append(p)
        return all_params

    def copy(self, freeze_parameters: bool = False) -> "PhotonicCircuit":
        """
        Creates and returns an identical copy of the circuit, optionally
        freezing parameter values.

        Args:

            freeze_parameters (bool, optional) : Used to control where any
                existing parameter objects are carried over to the newly
                created circuit, or if the current parameter values should be
                used. Defaults to False.

        Returns:

            PhotonicCircuit : A new PhotonicCircuit object with the same
                circuit configuration as the original object.

        """
        new_circ = PhotonicCircuit(self.n_modes)
        if not freeze_parameters:
            new_circ.__circuit_spec = copy(self.__circuit_spec)
        else:
            copied_spec = deepcopy(self.__circuit_spec)
            new_circ.__circuit_spec = list(self._freeze_params(copied_spec))
        new_circ.__in_heralds = copy(self.__in_heralds)
        new_circ.__out_heralds = copy(self.__out_heralds)
        new_circ.__external_in_heralds = copy(self.__external_in_heralds)
        new_circ.__external_out_heralds = copy(self.__external_out_heralds)
        new_circ.__internal_modes = copy(self.__internal_modes)
        return new_circ

    def unpack_groups(self) -> None:
        """
        Unpacks any component groups which may have been added to the circuit.
        """
        self.__internal_modes = []
        self.__external_in_heralds = self.__in_heralds
        self.__external_out_heralds = self.__out_heralds
        self.__circuit_spec = unpack_circuit_spec(self.__circuit_spec)

    def compress_mode_swaps(self) -> None:
        """
        Takes a provided circuit spec and will try to compress any more swaps
        such that the circuit length is reduced. Note that any components in a
        group will be ignored.
        """
        # Convert circuit spec and then assign to attribute
        new_spec = compress_mode_swaps(deepcopy(self.__circuit_spec))
        self.__circuit_spec = new_spec

    def remove_non_adjacent_bs(self) -> None:
        """
        Removes any beam splitters acting on non-adjacent modes by replacing
        with a mode swap and adjacent beam splitters.
        """
        # Convert circuit spec and then assign to attribute
        spec = deepcopy(self.__circuit_spec)
        new_spec = convert_non_adj_beamsplitters(spec)
        self.__circuit_spec = new_spec

    def _build(self) -> CompiledPhotonicCircuit:
        """
        Converts the ParameterizedCircuit into a circuit object using the
        components added and current values of the parameters.
        """
        try:
            circuit = self._build_process()
        except Exception as e:
            msg = "An error occurred during the circuit compilation process"
            raise CircuitCompilationError(msg) from e

        return circuit

    def _build_process(self) -> CompiledPhotonicCircuit:
        """
        Contains full process for convert a circuit into a compiled one.
        """
        circuit = CompiledPhotonicCircuit(self.n_modes)

        for spec in self.__circuit_spec:
            circuit.add(spec)

        heralds = self.heralds
        for i, o in zip(heralds.input, heralds.output, strict=True):
            circuit.add_herald(i, o, heralds.input[i], heralds.output[o])

        return circuit

    def _mode_in_range(self, mode: int) -> bool:
        """
        Check a mode exists within the created circuit and also confirm it is
        an integer.
        """
        if isinstance(mode, Parameter):
            raise TypeError("Mode values cannot be parameters.")
        # Catch this separately as bool is subclass of int
        if isinstance(mode, bool):
            raise TypeError("Mode number should be an integer.")
        if not isinstance(mode, int) and int(mode) != mode:
            raise TypeError("Mode number should be an integer.")
        if not (0 <= mode < self.n_modes):
            raise ModeRangeError(
                "Selected mode(s) is not within the range of the created "
                "circuit."
            )
        return True

    def _map_mode(self, mode: int) -> int:
        """
        Maps a provided mode to the corresponding internal mode
        """
        for i in sorted(self.__internal_modes):
            if mode >= i:
                mode += 1
        return mode

    def _add_empty_mode(self, mode: int) -> None:
        """
        Adds an empty mode at the selected location to a provided circuit spec.
        """
        self.__n_modes += 1
        self.__circuit_spec = add_empty_mode_to_circuit_spec(
            self.__circuit_spec, mode
        )
        # Also modify heralds as required
        to_modify = [
            "__in_heralds",
            "__out_heralds",
            "__external_in_heralds",
            "__external_out_heralds",
        ]
        for tm in to_modify:
            new_heralds = {}
            for m, n in getattr(self, "_PhotonicCircuit" + tm).items():
                m += 1 if m >= mode else 0  # noqa: PLW2901
                new_heralds[m] = n
            setattr(self, "_PhotonicCircuit" + tm, new_heralds)
        # Add internal mode storage
        self.__internal_modes = [
            m + 1 if m >= mode else m for m in self.__internal_modes
        ]

    def _freeze_params(self, circuit_spec: list[Component]) -> list[Component]:
        """
        Takes a provided circuit spec and will remove get any Parameter objects
        with their currently set values.
        """
        new_spec: list[Component] = []
        # Loop over spec and either call function again or add the value to the
        # new spec
        for spec in circuit_spec:
            spec = copy(spec)  # noqa: PLW2901
            if isinstance(spec, Group):
                spec.circuit_spec = self._freeze_params(spec.circuit_spec)
                new_spec.append(spec)
            else:
                for name, value in zip(
                    spec.fields(), spec.values(), strict=True
                ):
                    if isinstance(value, Parameter):
                        setattr(spec, name, value.get())
                    if isinstance(value, ParameterizedUnitary):
                        setattr(spec, name, value.unitary)
                new_spec.append(spec)
        return new_spec

    def _get_circuit_spec(self) -> list[Component]:
        """Returns a copy of the circuit spec attribute."""
        return deepcopy(self.__circuit_spec)



from setuptools import find_packages, setup

version = {}
with open("lightworks/__version.py") as f:
    exec(f.read(), version)

setup(
    name="lightworks",
    author="Aegiq Ltd.",
    version=version["__version__"],
    description="Open-source Python SDK for photonic quantum computation.",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Aegiq/lightworks",
    project_urls={
        "Source": "https://github.com/Aegiq/lightworks",
        "Documentation": "https://aegiq.github.io/lightworks/",
    },
    license="Apache 2.0",
    packages=find_packages(where=".", exclude=["tests"]),
    package_data={"lightworks": ["py.typed"]},
    python_requires=">=3.10",
    install_requires=[
        "thewalrus==0.20.0",
        "matplotlib>=3.7.1",
        "pandas>=2.0.1",
        "numpy>=1.24.3",
        "multimethod>=1.11.2",
        "drawsvg>=2.3.0",
        "CairoSVG>=2.8.0",
        "sympy>=1.12.0",
        "pyarrow",
        "ipython",
    ],
    extras_require={
        "all": ["lightworks[qiskit, remote]"],
        "remote": ["lightworks-remote>=1.2.0"],
        "qiskit": ["qiskit[visualization]>=1.1.0"],
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Operating System :: OS Independent",
    ],
)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from lightworks.interferometers import ErrorModel
from lightworks.interferometers.dists import (
    Distribution,
    Gaussian,
    TopHat,
)


class TestErrorModel:
    """
    Tests for Error Model object of module.
    """

    def test_default_bs_reflectivity(self):
        """
        Checks that the default beam splitter reflectivity is 0.5.
        """
        em = ErrorModel()
        # Repeat 100 times to confirm no randomness present
        for _i in range(100):
            assert em.get_bs_reflectivity() == 0.5

    def test_default_loss(self):
        """
        Checks that default loss value is 0.
        """
        em = ErrorModel()
        # Repeat 100 times to confirm no randomness present
        for _i in range(100):
            assert em.get_loss() == 0

    @pytest.mark.parametrize("dist", [Gaussian(0.5, 0.1), TopHat(0.1, 0.9)])
    def test_set_random_seed(self, dist):
        """
        Checks that set random seed produces identical values on subsequent
        calls when using each Distribution object with an rng.
        """
        em = ErrorModel()
        em.bs_reflectivity = dist
        # Set seed and get values twice
        em._set_random_seed(11)
        v1 = em.get_bs_reflectivity()
        em._set_random_seed(11)
        v2 = em.get_bs_reflectivity()
        # Check equivalence
        assert v1 == v2

    def test_set_random_seed_identical_distributions(self):
        """
        Checks that set random seed produces different values when two
        quantities are assigned to identical distributions.
        """
        em = ErrorModel()
        em.bs_reflectivity = Gaussian(0.5, 0.1)
        em.loss = Gaussian(0.5, 0.1)
        # Set seed and get values reflectivity and loss
        em._set_random_seed(11)
        v1 = em.get_bs_reflectivity()
        em._set_random_seed(11)
        v2 = em.get_loss()
        # Check equivalence
        assert v1 != v2

    def test_bs_reflectivity_is_distribution(self):
        """
        Checks that return by bs_reflectivity property is a Distribution object.
        """
        assert isinstance(ErrorModel().bs_reflectivity, Distribution)

    @pytest.mark.parametrize("value", [1, True, None, "Distribution"])
    def test_bs_reflectivity_enforces_distribution(self, value):
        """
        Checks that bs_reflectivity setting enforces that the value is a
        Distribution object.
        """
        em = ErrorModel()
        with pytest.raises(TypeError):
            em.bs_reflectivity = value

    def test_loss_is_distribution(self):
        """
        Checks that return by loss property is a Distribution object.
        """
        assert isinstance(ErrorModel().loss, Distribution)

    @pytest.mark.parametrize("value", [1, True, None, "Distribution"])
    def test_loss_enforces_distribution(self, value):
        """
        Checks that setting loss setting enforces that the value is a
        Distribution object.
        """
        em = ErrorModel()
        with pytest.raises(TypeError):
            em.loss = value

    def test_phase_offset_is_distribution(self):
        """
        Checks that return by phase_offset property is a Distribution object.
        """
        assert isinstance(ErrorModel().phase_offset, Distribution)

    @pytest.mark.parametrize("value", [1, True, None, "Distribution"])
    def test_phase_offset_enforces_distribution(self, value):
        """
        Checks that setting phase_offset setting enforces that the value is a
        Distribution object.
        """
        em = ErrorModel()
        with pytest.raises(TypeError):
            em.phase_offset = value

    @pytest.mark.parametrize("param", ["bs_reflectivity", "loss"])
    def test_parameters_in_string_and_repr(self, param):
        """
        Checks that all parameters of the error model are contained in the
        string and repr returns.
        """
        em = ErrorModel()
        assert param in str(em)
        assert param in repr(em)

    @pytest.mark.parametrize(
        "value", ["bs_reflectivity", "loss", "phase_offset"]
    )
    def test_random_seeding(self, value):
        """
        Checks that set random seed produces different values when two
        quantities are assigned to identical distributions.
        """
        em = ErrorModel()
        setattr(em, value, Gaussian(0.5, 0.1))
        # Set seed and get value twice
        em._set_random_seed(11)
        v1 = getattr(em, value).value()
        em._set_random_seed(11)
        v2 = getattr(em, value).value()
        # Check equivalence
        assert v1 == v2



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import copy
from dataclasses import dataclass
from typing import Any

import numpy as np

from lightworks.__settings import settings
from lightworks.emulator.components import Source
from lightworks.emulator.utils.exceptions import BackendError
from lightworks.sdk.tasks import SamplerTask, TaskData


@dataclass
class CacheData:
    """
    Stores cached data from the backend.

    Args:

        values: The relevant values used in calculating results for a particular
            cache.

        results: The calculated results dictionary.

    """

    values: list[Any]
    results: dict[Any, Any]


def check_parameter_updates(values1: list[Any], values2: list[Any]) -> bool:
    """
    Determines if parameters have changed between two sets of values.
    """
    for v1, v2 in zip(values1, values2, strict=True):
        # Treat arrays and other values differently
        if isinstance(v1, np.ndarray) and isinstance(v2, np.ndarray):
            if v1.shape != v2.shape:
                return True
            if not (v1 == v2).all():
                return True
        elif isinstance(v1, Source) and isinstance(v2, Source):
            if v1.brightness != v2.brightness:
                return True
            if v1.indistinguishability != v2.indistinguishability:
                return True
            if v1.purity != v2.purity:
                return True
            if v1.probability_threshold != v2.probability_threshold:
                return True
        elif v1 != v2:
            return True
    return False


def get_calculation_values(data: TaskData) -> list[Any]:
    """
    Stores all current parameters used with the sampler in a list and
    returns this.
    """
    if isinstance(data, SamplerTask):
        # Store all values which alter a computation
        return [
            data.circuit.U_full,
            data.circuit.heralds,
            data.input_state,
            copy(data.source),
            copy(settings.sampler_probability_threshold),
        ]
    raise BackendError("Caching not implemented for provided task type.")



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from numpy import arccos, pi

from lightworks import PhotonicCircuit, PostSelection, Sampler, State
from lightworks.emulator import Backend, Detector, Source
from lightworks.qubit import CNOT

heralds = [
    lambda s: (
        s[0] == 0 and s[5] == 0 and s[1] + s[2] == 1 and s[3] + s[4] == 1
    ),
    lambda s: (s[0] + s[1] == 1 and s[2] + s[3] == 1),
]
# Including heralds
post_selection = PostSelection()
post_selection.add(0, 0)
post_selection.add(5, 0)
post_selection.add((1, 2), 1)
post_selection.add((3, 4), 1)
# Excluding heralds
post_selection2 = PostSelection()
post_selection2.add((0, 1), 1)
post_selection2.add((2, 3), 1)

post_selects = [post_selection, post_selection2]


@pytest.mark.flaky(max_runs=3)
@pytest.mark.parametrize(
    ("backend", "post_selection"),
    [
        (Backend("permanent"), heralds),
        (Backend("slos"), heralds),
        (Backend("permanent"), post_selects),
        (Backend("slos"), post_selects),
    ],
)
class TestCNOT:
    """
    Samples from a non-heralded CNOT circuit with loss, imperfect source and
    detector to check that the correct results is produced. It tests this with
    both permanent and slos backends.
    """

    def test_cnot_sample_n_inputs(self, backend, post_selection):
        """
        Checks the correct output is produced from the CNOT gate when sampling
        N inputs from the system. Note, very occasionally this test may fail
        due to the probabilistic nature of the sampler. It is only a concern if
        this happens consistently.
        """
        # Create CNOT circuit
        r = 1 / 3
        loss = 0.1
        theta = arccos(r)
        cnot_circuit = PhotonicCircuit(6)
        to_add = [
            (3, pi / 2, 0),
            (0, theta, 0),
            (2, theta, pi),
            (4, theta, 0),
            (3, pi / 2, 0),
        ]
        for m, t, p in to_add:
            cnot_circuit.bs(m, loss=loss, reflectivity=0.5)
            cnot_circuit.ps(m + 1, t)
            cnot_circuit.bs(m, loss=loss, reflectivity=0.5)
            cnot_circuit.ps(m + 1, p)
            if m in {3, 4}:
                cnot_circuit.barrier()
        # Define imperfect source and detector
        source = Source(purity=0.99, brightness=0.4, indistinguishability=0.96)
        detector = Detector(efficiency=0.9, p_dark=1e-5, photon_counting=False)
        # Then define sampler with the input state |10>
        sampler = Sampler(
            cnot_circuit,
            State([0, 0, 1, 1, 0, 0]),
            20000,
            source=source,
            detector=detector,
            post_selection=post_selection[0],
            sampling_mode="input",
        )
        results = backend.run(sampler)
        # We expect the state |11> (|0,0,1,0,1,0> in mode language) with
        # reasonable fidelity, so we will assert this is measured for > 80% of
        # the total samples which met the herald condition
        eff = results[State([0, 0, 1, 0, 1, 0])] / sum(results.values())
        assert eff > 0.8

    def test_cnot_sample_n_outputs(self, backend, post_selection):
        """
        Checks the correct output is produced from the CNOT gate when sampling
        N outputs from the system. Note, very occasionally this test may fail
        due to the probabilistic nature of the sampler. It is only a concern if
        this happens consistently.
        """
        # Create CNOT circuit
        r = 1 / 3
        loss = 0.1
        theta = arccos(r)
        cnot_circuit = PhotonicCircuit(6)
        to_add = [
            (3, pi / 2, 0),
            (0, theta, 0),
            (2, theta, pi),
            (4, theta, 0),
            (3, pi / 2, 0),
        ]
        for m, t, p in to_add:
            cnot_circuit.bs(m, loss=loss, reflectivity=0.5)
            cnot_circuit.ps(m + 1, t)
            cnot_circuit.bs(m, loss=loss, reflectivity=0.5)
            cnot_circuit.ps(m + 1, p)
            if m in {3, 4}:
                cnot_circuit.barrier()
        # Define imperfect source and detector
        source = Source(purity=0.99, brightness=0.4, indistinguishability=0.94)
        detector = Detector(efficiency=1, photon_counting=False)
        # Then define sampler with the input state |10>
        sampler = Sampler(
            cnot_circuit,
            State([0, 0, 1, 1, 0, 0]),
            20000,
            source=source,
            detector=detector,
            post_selection=post_selection[0],
        )
        results = backend.run(sampler)
        # We expect the state |11> (|0,0,1,0,1,0> in mode language) with
        # reasonable fidelity, so we will assert this is measured for > 80% of
        # the total samples which met the herald condition
        eff = results[State([0, 0, 1, 0, 1, 0])] / 20000
        assert eff > 0.8

    def test_cnot_sample_n_inputs_built_in(self, backend, post_selection):
        """
        Checks the correct output is produced from the built-in CNOT gate when
        sampling N inputs from the system. Note, very occasionally this test
        may fail due to the probabilistic nature of the sampler. It is only a
        concern if this happens consistently.
        """
        # Create CNOT circuit
        cnot_circuit = CNOT()
        # Define imperfect source and detector
        source = Source(purity=0.99, brightness=0.4, indistinguishability=0.94)
        detector = Detector(efficiency=0.9, p_dark=1e-5, photon_counting=False)
        # Then define sampler with the input state |10>
        sampler = Sampler(
            cnot_circuit,
            State([0, 1, 1, 0]),
            20000,
            source=source,
            detector=detector,
            post_selection=post_selection[1],
            sampling_mode="input",
        )
        results = backend.run(sampler)
        # We expect the state |11> (|0,1,0,1> in mode language) with
        # reasonable fidelity, so we will assert this is measured for > 80% of
        # the total samples which met the herald condition
        eff = results[State([0, 1, 0, 1])] / sum(results.values())
        assert eff > 0.8

    def test_cnot_sample_n_outputs_built_in(self, backend, post_selection):
        """
        Checks the correct output is produced from the built-in CNOT gate when
        sampling N outputs from the system. Note, very occasionally this test
        may fail due to the probabilistic nature of the sampler. It is only a
        concern if this happens consistently.
        """
        # Create CNOT circuit
        cnot_circuit = CNOT()
        # Define imperfect source and detector
        source = Source(purity=0.99, brightness=0.4, indistinguishability=0.94)
        detector = Detector(efficiency=1, photon_counting=False)
        # Then define sampler with the input state |10>
        sampler = Sampler(
            cnot_circuit,
            State([0, 1, 1, 0]),
            20000,
            source=source,
            detector=detector,
            post_selection=post_selection[1],
        )
        results = backend.run(sampler)
        # We expect the state |11> (|0,1,0,1> in mode language) with
        # reasonable fidelity, so we will assert this is measured for > 80% of
        # the total samples which met the herald condition
        eff = results[State([0, 1, 0, 1])] / 20000
        assert eff > 0.8



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from math import factorial, prod

import numpy as np
from numpy.typing import NDArray
from thewalrus import perm

from lightworks.__settings import settings
from lightworks.emulator.utils.state import fock_basis
from lightworks.sdk.circuit.photonic_compiler import CompiledPhotonicCircuit
from lightworks.sdk.state import State

from .fock_backend import FockBackend


class PermanentBackend(FockBackend):
    """
    Calculate the permanent for a give unitary matrix and input state. In this
    case, thewalrus module is used for all permanent calculations.
    """

    @property
    def name(self) -> str:
        """Returns the name of the backend"""
        return "permanent"

    @property
    def compatible_tasks(self) -> tuple[str, ...]:
        """Returns backends which are compatible with the backend."""
        return ("Sampler", "Analyzer", "Simulator")

    def state_generator(self, n_modes: int, n_photons: int) -> list[list[int]]:
        """
        Generates all possible photonic states for a given number of modes and
        photons.
        """
        return fock_basis(n_modes, n_photons)

    def probability_amplitude(
        self,
        unitary: NDArray[np.complex128],
        input_state: list[int],
        output_state: list[int],
    ) -> complex:
        """
        Find the probability amplitude between a given input and output state
        for the provided unitary. Note values should be provided as an
        array/list.

        Args:

            unitary (np.ndarray) : The target unitary matrix which represents
                the transformation implemented by a circuit.

            input_state (list) : The input state to the system.

            output_state (list) : The target output state.

        Returns:

            complex : The calculated probability amplitude.

        """
        factor_m = prod([factorial(i) for i in input_state])
        factor_n = prod([factorial(i) for i in output_state])
        # Calculate permanent for given input/output
        return perm(partition(unitary, input_state, output_state)) / (
            np.sqrt(factor_m * factor_n)
        )

    def probability(
        self,
        unitary: NDArray[np.complex128],
        input_state: list[int],
        output_state: list[int],
    ) -> float:
        """
        Calculates the probability of a given output state for a provided
        unitary and input state to the system. Note values should be provided
        as an array/list.

        Args:

            unitary (np.ndarray) : The target unitary matrix which represents
                the transformation implemented by a circuit.

            input_state (list) : The input state to the system.

            output_state (list) : The target output state.

        Returns:

            float : The calculated probability of transition between the input
                and output.

        """
        return (
            abs(self.probability_amplitude(unitary, input_state, output_state))
            ** 2
        )

    def full_probability_distribution(
        self, circuit: CompiledPhotonicCircuit, input_state: State
    ) -> dict[State, float]:
        """
        Finds the output probability distribution for the provided circuit and
        input state.

        Args:

            circuit (CompiledPhotonicCircuit) : The compiled version of the
                circuit which is being simulated. This is created by calling the
                _build method on the target circuit.

            input_state (State) : The input state to the system.

        Returns:

            dict : The calculated full probability distribution for the input.

        """
        # Return empty distribution when 0 photons in input
        if input_state.n_photons == 0:
            return {State([0] * circuit.n_modes): 1.0}

        pdist: dict[State, float] = {}
        # Add extra states for loss modes here when included
        if circuit.loss_modes > 0:
            input_state += State([0] * circuit.loss_modes)
        # For a given input work out all possible outputs
        out_states = fock_basis(len(input_state), input_state.n_photons)
        for ostate in out_states:
            # Skip any zero photon states
            if sum(ostate[: circuit.n_modes]) == 0:
                continue
            p = self.probability(circuit.U_full, input_state.s, ostate)
            if p > settings.sampler_probability_threshold:
                # Only care about non-loss modes
                meas_state = State(ostate[: circuit.n_modes])
                if meas_state in pdist:
                    pdist[meas_state] += p
                else:
                    pdist[meas_state] = p
        # Work out zero photon component before saving to unique results
        total_prob = sum(pdist.values())
        if total_prob < 1 and circuit.loss_modes > 0:
            pdist[State([0] * circuit.n_modes)] = 1 - total_prob

        return pdist


def partition(
    unitary: NDArray[np.complex128], in_state: list[int], out_state: list[int]
) -> NDArray[np.complex128]:
    """
    Converts the unitary matrix into a larger matrix used for in the
    permanent calculation.
    """
    n_modes = len(in_state)  # Number of modes
    # Construct the matrix of indices for the partition
    x = [i for i in range(n_modes) for _ in range(out_state[i])]
    y = [i for i in range(n_modes) for _ in range(in_state[i])]
    # Construct the new matrix with dimension n, where n is photon number
    return unitary[np.ix_(x, y)]



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains all global settings for use across lightworks. Any new values
should be added the settings class below.
"""

from typing import Any


class _Settings:
    """
    Stores all global settings for Lightworks. After creation new attributes
    (settings) cannot be added.
    """

    __frozen: bool = False
    unitary_precision: float = 1e-10
    sampler_probability_threshold: float = 1e-9

    def __init__(self) -> None:
        self.__frozen = True

    @property
    def all(self) -> list[str]:
        return ["unitary_precision", "sampler_probability_threshold"]

    def __str__(self) -> str:
        output = ""
        for val in self.all:
            output += f"{val}: {getattr(self, val)}\n"
        return output[:-1]

    def __repr__(self) -> str:
        return f"lightworks.Settings(\n{self!s}\n)"

    def __setattr__(self, name: str, value: Any) -> None:
        if not hasattr(self, name) and self.__frozen:
            msg = f"Setting with name '{name}' does not exist."
            raise AttributeError(msg)
        super().__setattr__(name, value)

    def __getattr__(self, name: str) -> Any:
        # This is only called when attribute doesn't exist, so can use this to
        # raise a custom exception
        msg = f"Setting with name '{name}' does not exist."
        raise AttributeError(msg)


settings = _Settings()



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from random import randint, random, seed

import pytest
import sympy as sp

from lightworks import (
    CircuitCompilationError,
    LightworksDependencyError,
    Parameter,
    ParameterDict,
    PhotonicCircuit,
    Unitary,
    convert,
    random_unitary,
)
from lightworks.qubit import CNOT, Rx
from lightworks.sdk.circuit.photonic_components import (
    BeamSplitter,
    Component,
    Group,
    ModeSwaps,
)
from lightworks.sdk.utils.param_unitary import ParameterizedUnitary


class TestCircuit:
    """
    Unit tests to confirm correct functioning of the PhotonicCircuit class when
    various operations are performed.
    """

    def setup_method(self) -> None:
        """Create a circuit and associated parameters for testing."""
        n_modes = 6
        self.param_circuit = PhotonicCircuit(n_modes)
        self.parameters = ParameterDict()
        self.original_parameters = ParameterDict()
        seed(1)
        for i in range(n_modes - 1):
            for j in range(i % 2, n_modes - i % 2, 2):
                p1 = Parameter(random())
                p1c = Parameter(p1.get())
                p2 = Parameter(random())
                p2c = Parameter(p2.get())
                self.parameters[f"bs_{i}_{j}"] = p1
                self.parameters[f"ps_{i}_{j}"] = p2
                self.original_parameters[f"bs_{i}_{j}"] = p1c
                self.original_parameters[f"ps_{i}_{j}"] = p2c
                self.param_circuit.ps(j, self.parameters[f"ps_{i}_{j}"])
                self.param_circuit.bs(j)
                self.param_circuit.ps(j + 1, self.parameters[f"bs_{i}_{j}"])
                self.param_circuit.bs(j, loss=convert.db_loss_to_decimal(0.1))

    def test_resultant_unitary(self):
        """
        Checks that the resultant unitary from a parameterized circuit is as
        expected.
        """
        unitary = self.param_circuit.U
        assert unitary[0, 0] == pytest.approx(
            0.1817783235792 + 0.261054657406j, 1e-8
        )
        assert unitary[1, 2] == pytest.approx(
            0.1094958407210 - 0.2882179078302j, 1e-8
        )
        assert unitary[4, 3] == pytest.approx(
            0.03978296812819 + 0.354080300183j, 1e-8
        )

    def test_parameter_modification(self):
        """
        Confirms that parameter modification changes unitary in expected way.
        """
        self.parameters["bs_0_0"] = 4
        self.parameters["bs_0_2"] = 4
        unitary = self.param_circuit.U
        assert unitary[0, 0] == pytest.approx(
            0.1382843851268 - 0.1276219199576j, 1e-8
        )
        assert unitary[1, 2] == pytest.approx(
            0.6893944687270 + 0.2987967171732j, 1e-8
        )
        assert unitary[4, 3] == pytest.approx(
            -0.82752490939 - 0.0051178352488j, 1e-8
        )

    def test_circuit_addition(self):
        """Confirms two circuits are added together correctly."""
        new_circ = self.param_circuit + self.param_circuit
        unitary = new_circ.U
        assert unitary[0, 0] == pytest.approx(
            0.2743757510982 + 0.6727464244294j, 1e-8
        )
        assert unitary[1, 2] == pytest.approx(
            -0.153884469732 + 0.0872489579891j, 1e-8
        )
        assert unitary[4, 3] == pytest.approx(
            -0.083445311860 + 0.154159863276j, 1e-8
        )

    def test_smaller_circuit_addition(self):
        """
        Confirms equivalence between building a single circuit and added a
        larger circuit to a smaller one with the add method.
        """
        # Comparison circuit
        circ_comp = PhotonicCircuit(6)
        # First part
        for i, m in enumerate([0, 2, 4, 1, 3, 2]):
            circ_comp.bs(m)
            circ_comp.ps(m, i)
        # Second part
        for i, m in enumerate([3, 1, 3, 2, 1]):
            circ_comp.ps(m + 1, i)
            circ_comp.bs(m, loss=0.2 * i)
            circ_comp.ps(m, i, loss=0.1)
        # Addition circuit
        c1 = PhotonicCircuit(6)
        for i, m in enumerate([0, 2, 4, 1, 3, 2]):
            c1.bs(m)
            c1.ps(m, i)
        c2 = PhotonicCircuit(4)
        for i, m in enumerate([2, 0, 2, 1, 0]):
            c2.ps(m + 1, i)
            c2.bs(m, loss=0.2 * i)
            c2.ps(m, i, loss=0.1)
        c1.add(c2, 1)
        # Check unitary equivalence
        u_1 = circ_comp.U_full.round(8)
        u_2 = c1.U_full.round(8)
        assert (u_1 == u_2).all()

    def test_smaller_circuit_addition_grouped(self):
        """
        Confirms equivalence between building a single circuit and added a
        larger circuit to a smaller one with the add method, while using the
        group method.
        """
        # Comparison circuit
        circ_comp = PhotonicCircuit(6)
        # First part
        for i, m in enumerate([0, 2, 4, 1, 3, 2]):
            circ_comp.bs(m)
            circ_comp.ps(m, i)
        # Second part
        for _i in range(4):
            for i, m in enumerate([3, 1, 3, 2, 1]):
                circ_comp.ps(m + 1, i)
                circ_comp.bs(m, loss=0.2 * i)
                circ_comp.ps(m, i, loss=0.1)
            circ_comp.mode_swaps({1: 2, 2: 3, 3: 1})
        # Addition circuit
        c1 = PhotonicCircuit(6)
        for i, m in enumerate([0, 2, 4, 1, 3, 2]):
            c1.bs(m)
            c1.ps(m, i)
        c2 = PhotonicCircuit(4)
        for i, m in enumerate([2, 0, 2, 1, 0]):
            c2.ps(m + 1, i)
            c2.bs(m, loss=0.2 * i)
            c2.ps(m, i, loss=0.1)
        c2.mode_swaps({0: 1, 1: 2, 2: 0})
        # Test combinations of True and False for group option
        c2.add(c2, 0, group=True)
        c2.add(c2, 0, group=False)
        c1.add(c2, 1, group=True)
        # Check unitary equivalence
        u_1 = circ_comp.U_full.round(8)
        u_2 = c1.U_full.round(8)
        assert (u_1 == u_2).all()

    @pytest.mark.parametrize("herald", [True, False])
    def test_circuit_add_doesnt_modify(self, herald):
        """
        Checks that the circuit add doesn't modify the circuit being added.
        """
        c1 = PhotonicCircuit(6)
        # Create heralded circuit and add
        c2 = PhotonicCircuit(3)
        c2.herald(1, 1)
        c1.add(c2)
        # Then add non-heralded circuit to that.
        c3 = PhotonicCircuit(4)
        c3.bs(0)
        c3.bs(1)
        c3.bs(2)
        if herald:
            c3.herald(1, 1)
        before_u = c3.U_full  # Save unitary
        c1.add(c3)
        # Check unitary preserved
        assert c3.n_modes == 4
        assert (before_u == c3.U_full).all()

    def test_circuit_add_doesnt_modify_group(self):
        """
        Checks that the circuit add doesn't modify the circuit being added when
        a circuit is grouped manuallys.
        """
        c1 = PhotonicCircuit(6)
        # Create heralded circuit and add
        c2 = PhotonicCircuit(3)
        c2.herald(1, 1)
        c1.add(c2)
        # Then add non-heralded circuit to that.
        c3 = PhotonicCircuit(4)
        c3.bs(0)
        c3.bs(1)
        c3.herald(1, 1)
        c3.bs(2)
        before_u = c3.U_full  # Save unitary
        c1.add(c3, 0, group=True)
        # Check unitary preserved
        assert c3.n_modes == 4
        assert (before_u == c3.U_full).all()

    def test_barrier_inclusion(self):
        """
        Checks that barrier component can be added across all and a selected
        mode range.
        """
        circuit = PhotonicCircuit(4)
        circuit.barrier()
        circuit.barrier([0, 2])

    def test_mode_not_parameter(self):
        """
        Checks that an error is raised if a parameter is attempted to be
        assigned to a mode value.
        """
        new_circ = PhotonicCircuit(4)
        with pytest.raises(TypeError):
            new_circ.bs(Parameter(1))
        with pytest.raises(TypeError):
            new_circ.ps(Parameter(1))
        with pytest.raises(TypeError):
            new_circ.mode_swaps({Parameter(1): 2, 2: Parameter(1)})

    def test_circ_unitary_combination(self):
        """Test combination of a circuit and unitary objects."""
        circ = PhotonicCircuit(6)
        for i, m in enumerate([0, 2, 4, 1, 3, 2]):
            circ.bs(m, loss=convert.db_loss_to_decimal(0.2))
            circ.ps(m, i)
        u1 = Unitary(random_unitary(6, seed=1))
        u2 = Unitary(random_unitary(4, seed=2))
        circ.add(u1, 0)
        circ.add(u2, 1)
        assert circ.U[0, 0] == pytest.approx(
            0.2287112952348 - 0.14731470234581j, 1e-8
        )
        assert circ.U[1, 2] == pytest.approx(
            0.0474053983616 + 0.01248244201229j, 1e-8
        )
        assert circ.U[4, 3] == pytest.approx(
            0.0267553699139 - 0.02848937675632j, 1e-8
        )

    def test_mode_modification(self):
        """
        Checks that n_modes attribute cannot be modified and will raise an
        attribute error.
        """
        circ = PhotonicCircuit(4)
        with pytest.raises(AttributeError):
            circ.n_modes = 6

    def test_circuit_copy(self):
        """Test copy method of circuit creates an independent circuit."""
        copied_circ = self.param_circuit.copy()
        u_1 = self.param_circuit.U_full
        # Modify the new circuit and check the original U is unchanged
        copied_circ.bs(0)
        u_2 = self.param_circuit.U_full
        assert (u_1 == u_2).all()

    def test_circuit_copy_parameter_modification(self):
        """Test parameter modification still works on a copied circuit"""
        copied_circ = self.param_circuit.copy()
        u_1 = copied_circ.U_full
        # Modify parameter and get new unitary from copied circuit
        self.parameters["bs_0_0"] = 2
        u_2 = copied_circ.U_full
        # Unitary should be modified
        assert not (u_1 == u_2).all()

    def test_circuit_copy_parameter_freeze(self):
        """
        Checks copy method of the circuit can be used with the freeze parameter
        argument to create a new circuit without the Parameter objects.
        """
        copied_circ = self.param_circuit.copy(freeze_parameters=True)
        u_1 = copied_circ.U_full
        # Modify parameter and get new unitary from copied circuit
        self.parameters["bs_0_0"] = 4
        u_2 = copied_circ.U_full
        # Unitary should not be modified
        assert (u_1 == u_2).all()

    def test_parameterized_unitary_copy_parameter_freeze(self):
        """
        Checks copy method of the circuit can be used with the freeze parameter
        argument to create a new circuit without the Parameter objects, when a
        parameterized unitary is used.
        """
        param = Parameter(1)
        circ = Rx(theta=param)
        copied_circ = circ.copy(freeze_parameters=True)
        u_1 = copied_circ.U_full
        # Modify parameter and get new unitary from copied circuit
        param.set(2)
        u_2 = copied_circ.U_full
        # Unitary should not be modified
        assert (u_1 == u_2).all()

    def test_circuit_copy_parameter_freeze_group(self):
        """
        Checks copy method of the circuit can be used with the freeze parameter
        argument to create a new circuit without the Parameter objects, while
        one of the parameters is in a group.
        """
        circ = PhotonicCircuit(self.param_circuit.n_modes + 2)
        circ.bs(0)
        circ.bs(2)
        circ.add(self.param_circuit, 1, group=True)
        copied_circ = circ.copy(freeze_parameters=True)
        u_1 = copied_circ.U_full
        # Modify parameter and get new unitary from copied circuit
        self.parameters["bs_0_0"] = 4
        u_2 = copied_circ.U_full
        # Unitary should not be modified
        assert (u_1 == u_2).all()

    def test_circuit_copy_preserves_heralds(self):
        """
        Tests that circuit copy method is able to correctly copy over the
        heralding data.
        """
        circ = CNOT()
        circ.add(CNOT(), 0)
        circ_copy = circ.copy()
        # Check circuit has heralds and that they are identical
        assert circ_copy.heralds
        assert circ.heralds == circ_copy.heralds

    def test_circuit_copy_preserves_internal_modes(self):
        """
        Tests that circuit copy method is able to correctly copy over the
        internal mode data used for storing heralds.
        """
        circ = CNOT()
        circ.add(CNOT(), 0)
        circ_copy = circ.copy()
        # Check circuit has internal modes and that they are preserved
        assert circ_copy._PhotonicCircuit__internal_modes
        assert (
            circ._PhotonicCircuit__internal_modes
            == circ_copy._PhotonicCircuit__internal_modes
        )

    def test_circuit_ungroup(self):
        """
        Check that the unpack_groups method removes any grouped components from
        the circuit.
        """
        # Create initial basic circuit
        circuit = PhotonicCircuit(4)
        circuit.bs(0)
        circuit.bs(2)
        circuit.ps(1, 0)
        # Create smaller circuit to add and combine
        circuit2 = PhotonicCircuit(2)
        circuit2.bs(0)
        circuit2.ps(1, 1)
        circuit2.add(circuit2, group=True)
        circuit.add(circuit2, 1, group=True)
        # Apply unpacking and check any groups have been removed
        circuit.unpack_groups()
        group_found = False
        for spec in circuit._get_circuit_spec():
            if isinstance(spec, Group):
                group_found = True
        assert not group_found

    def test_remove_non_adj_bs_success(self):
        """
        Checks that the remove_non_adjacent_bs method of the circuit is able
        to successfully remove all beam splitters which act on non-adjacent
        modes.
        """
        # Create circuit with beam splitters across non-adjacent modes
        circuit = PhotonicCircuit(8)
        circuit.bs(0, 7)
        circuit.bs(1, 4)
        circuit.bs(2, 6)
        circuit.bs(3, 7)
        circuit.bs(0, 1)
        # Apply method and check all remaining beam splitters
        circuit.remove_non_adjacent_bs()
        for spec in circuit._get_circuit_spec():
            # Check it acts on adjacent modes, otherwise fail
            if (
                isinstance(spec, BeamSplitter)
                and spec.mode_1 != spec.mode_2 - 1
            ):
                pytest.fail(
                    "Beam splitter which acts on non-adjacent modes found "
                    "in circuit spec."
                )

    def test_remove_non_adj_bs_grouped_success(self):
        """
        Checks that the remove_non_adjacent_bs method of the circuit is able
        to successfully remove all beam splitters which act on non-adjacent
        modes when groups are included.
        """
        # Create circuit with beam splitters across non-adjacent modes
        circuit = PhotonicCircuit(8)
        circuit.bs(0, 7)
        circuit.bs(1, 4)
        circuit.bs(2, 6)
        circuit.bs(3, 7)
        circuit.bs(0, 1)
        # Then create smaller second circuit and add, with group = True
        circuit2 = PhotonicCircuit(6)
        circuit2.bs(1, 4)
        circuit2.bs(2, 5)
        circuit2.bs(3)
        circuit.add(circuit2, 1, group=True)
        # Apply method and check all remaining beam splitters
        circuit.remove_non_adjacent_bs()
        circuit.unpack_groups()
        for spec in circuit._get_circuit_spec():
            # Check it acts on adjacent modes, otherwise fail
            if (
                isinstance(spec, BeamSplitter)
                and spec.mode_1 != spec.mode_2 - 1
            ):
                pytest.fail(
                    "Beam splitter which acts on non-adjacent modes found "
                    "in circuit spec."
                )

    def test_remove_non_adj_bs_equivalence(self):
        """
        Checks that the remove_non_adjacent_bs method of the circuit retains
        the circuit unitary.
        """
        # Create circuit with beam splitters across non-adjacent modes
        circuit = PhotonicCircuit(8)
        circuit.bs(0, 7)
        circuit.bs(1, 4)
        circuit.bs(2, 6)
        circuit.bs(7, 3)
        circuit.bs(0, 1)
        # Apply method and check unitary equivalence
        u1 = abs(circuit.U).round(3)
        circuit.remove_non_adjacent_bs()
        u2 = abs(circuit.U).round(3)
        assert (u1 == u2).all()

    def test_remove_non_adj_bs_equivalence_grouped(self):
        """
        Checks that the remove_non_adjacent_bs method of the circuit retains
        the circuit unitary when grouped components are featured in the
        circuit.
        """
        # Create circuit with beam splitters across non-adjacent modes
        circuit = PhotonicCircuit(8)
        circuit.bs(0, 7)
        circuit.bs(1, 4)
        circuit.bs(2, 6)
        circuit.bs(3, 7)
        circuit.bs(0, 1)
        # Then create smaller second circuit and add, with group = True
        circuit2 = PhotonicCircuit(6)
        circuit2.bs(1, 4)
        circuit2.bs(2, 5)
        circuit2.bs(3)
        circuit.add(circuit2, 1, group=True)
        # Apply method and check unitary equivalence
        u1 = abs(circuit.U).round(8)
        circuit.remove_non_adjacent_bs()
        u2 = abs(circuit.U).round(8)
        assert (u1 == u2).all()

    def test_compress_mode_swap_equivalance(self):
        """
        Tests the circuit compress_mode_swaps method retains the circuit
        unitary.
        """
        # Create circuit with a few components and then mode swaps
        circuit = PhotonicCircuit(8)
        circuit.bs(0)
        circuit.bs(4)
        circuit.mode_swaps({1: 3, 3: 5, 5: 6, 6: 1})
        circuit.mode_swaps({0: 1, 2: 4, 1: 2, 4: 0})
        circuit.mode_swaps({5: 3, 3: 5})
        circuit.bs(0)
        circuit.bs(4)
        # Apply method and check unitary equivalence
        u1 = abs(circuit.U).round(8)
        circuit.compress_mode_swaps()
        u2 = abs(circuit.U).round(8)
        assert (u1 == u2).all()

    def test_compress_mode_swap_equivalance_unitary(self):
        """
        Tests the circuit compress_mode_swaps method retains the circuit
        unitary while a unitary matrix component is included.
        """
        # Create circuit with a few components and then mode swaps
        circuit = PhotonicCircuit(8)
        circuit.bs(0)
        circuit.bs(4)
        circuit.add(Unitary(random_unitary(4)), 2)
        circuit.mode_swaps({1: 3, 3: 5, 5: 6, 6: 1})
        circuit.mode_swaps({0: 1, 2: 4, 1: 2, 4: 0})
        circuit.mode_swaps({5: 3, 3: 5})
        circuit.bs(0)
        circuit.bs(4)
        # Apply method and check unitary equivalence
        u1 = abs(circuit.U).round(8)
        circuit.compress_mode_swaps()
        u2 = abs(circuit.U).round(8)
        assert (u1 == u2).all()

    def test_compress_mode_swap_removes_components(self):
        """
        Tests the circuit compress_mode_swaps method is able to reduce it down
        to using 2 mode swaps for an example circuit.
        """
        # Create circuit with a few components and then mode swaps
        circuit = PhotonicCircuit(8)
        circuit.bs(0)
        circuit.bs(4)
        circuit.mode_swaps({1: 3, 3: 5, 5: 6, 6: 1})
        circuit.mode_swaps({0: 1, 2: 4, 1: 2, 4: 0})
        circuit.mode_swaps({5: 3, 3: 5})
        circuit.bs(0)
        circuit.bs(4)
        circuit.mode_swaps({0: 1, 2: 4, 1: 2, 4: 0})
        # Apply method and check only two mode_swap components are present
        circuit.compress_mode_swaps()
        counter = 0
        for spec in circuit._get_circuit_spec():
            if isinstance(spec, ModeSwaps):
                counter += 1
        assert counter == 2

    @pytest.mark.parametrize(
        ("component", "args"),
        [("bs", [2, 4]), ("ps", [4, 1]), ("loss", [4, 1])],
    )
    def test_compress_mode_swap_blocked_by_component(self, component, args):
        """
        Tests the circuit compress_mode_swaps method is blocked by a range of
        components.
        """
        # Create circuit with a few components and then mode swaps
        circuit = PhotonicCircuit(8)
        circuit.mode_swaps({4: 5, 5: 6, 6: 4})
        getattr(circuit, component)(*args)
        circuit.mode_swaps({4: 5, 5: 6, 6: 4})
        # Apply method and check two mode_swap components are still present
        circuit.compress_mode_swaps()
        counter = 0
        for spec in circuit._get_circuit_spec():
            if isinstance(spec, ModeSwaps):
                counter += 1
        assert counter == 2

    @pytest.mark.parametrize("mode", [1, 3, 5])
    def test_compress_mode_swap_blocked_by_unitary(self, mode):
        """
        Tests the circuit compress_mode_swaps method is blocked by a unitary
        matrix
        """
        # Create circuit with a few components and then mode swaps
        circuit = PhotonicCircuit(8)
        circuit.mode_swaps({3: 4, 4: 5, 5: 3})
        circuit.add(Unitary(random_unitary(3)), mode)
        circuit.mode_swaps({3: 4, 4: 5, 5: 3})
        # Apply method and check two mode_swap components are still present
        circuit.compress_mode_swaps()
        counter = 0
        for spec in circuit._get_circuit_spec():
            if isinstance(spec, ModeSwaps):
                counter += 1
        assert counter == 2

    def test_compress_mode_swap_ignores_groups(self):
        """Checks that the mode swap ignores components in groups."""
        # Create circuit with a few components and then two mode swaps, one
        # placed within a group from a smaller circuit
        circuit = PhotonicCircuit(8)
        circuit.bs(0)
        circuit.bs(4)
        circuit.mode_swaps({1: 3, 3: 5, 5: 6, 6: 1})
        circuit2 = PhotonicCircuit(5)
        circuit2.mode_swaps({0: 1, 2: 4, 1: 2, 4: 0})
        circuit.add(circuit2, 1, group=True)
        circuit.bs(0)
        circuit.bs(4)
        # Apply method and check two mode_swap components are still present
        circuit.compress_mode_swaps()
        circuit.unpack_groups()  # unpack groups for counting
        counter = 0
        for spec in circuit._get_circuit_spec():
            if isinstance(spec, ModeSwaps):
                counter += 1
        assert counter == 2

    def test_parameterized_loss(self):
        """Checks that loss can be parameterized in a circuit."""
        param = Parameter(0.5, label="loss")
        circuit = PhotonicCircuit(4)
        circuit.bs(0, loss=param)
        circuit.ps(2, 1, loss=param)
        circuit.loss(2, param)

    def test_herald(self):
        """
        Confirms that heralding being added to a circuit works as expected and
        is reflected in the heralds attribute.
        """
        circuit = PhotonicCircuit(4)
        circuit.herald((0, 2), (1, 3))
        # Check heralds added
        assert 0 in circuit.heralds.input
        assert 2 in circuit.heralds.output
        # Check photon number is correct
        assert circuit.heralds.input[0] == 1
        assert circuit.heralds.output[2] == 3

    def test_herald_single_value(self):
        """
        Confirms that heralding being added to a circuit works as expected and
        is reflected in the heralds attribute, when only a single mode is set
        """
        circuit = PhotonicCircuit(4)
        circuit.herald(1, 2)
        # Check heralds added
        assert 1 in circuit.heralds.input
        assert 1 in circuit.heralds.output
        # Check photon number is correct
        assert circuit.heralds.input[1] == 2
        assert circuit.heralds.output[1] == 2

    @pytest.mark.parametrize("value", [2.5, "2", True])
    def test_herald_invalid_photon_number(self, value):
        """
        Checks error is raised when a non-integer photon number is provided to
        the herald method.
        """
        circuit = PhotonicCircuit(4)
        with pytest.raises(TypeError):
            circuit.herald(0, value)

    @pytest.mark.parametrize("modes", [1, 2, (1, 2)])
    def test_herald_duplicate(self, modes):
        """
        Checks error is raised when duplicate mode is added to herald.
        """
        circuit = PhotonicCircuit(4)
        circuit.herald((1, 2), 1)
        with pytest.raises(ValueError):
            circuit.herald(modes, 1)

    def test_heralded_circuit_addition(self):
        """
        Checks that the heralds end up on the correct modes when a heralded
        circuit is added to a larger circuit.
        """
        circuit = PhotonicCircuit(4)
        # Create heralded sub-circuit to add
        sub_circ = PhotonicCircuit(4)
        sub_circ.herald((0, 0), 1)
        sub_circ.herald((3, 3), 1)
        # Then add to larger circuit
        circuit.add(sub_circ, 1)
        # Check circuit size is increased
        assert circuit.n_modes == 6
        # Confirm heralds are on modes 1 and 4
        assert 1 in circuit.heralds.input
        assert 1 in circuit.heralds.output
        assert 4 in circuit.heralds.input
        assert 4 in circuit.heralds.output

    def test_heralded_circuit_addition_herald_modification(self):
        """
        Checks that heralds in an existing circuit are modified correctly when
        a heralded sub-circuit is added.
        """
        # Place beam splitters across 0 & 1 and 2 & 3
        circuit = PhotonicCircuit(4)
        circuit.herald((0, 1), 0)
        circuit.herald((2, 3), 2)
        # Create heralded sub-circuit and add
        sub_circ = PhotonicCircuit(4)
        sub_circ.herald((0, 0), 1)
        sub_circ.herald((3, 3), 1)
        circuit.add(sub_circ, 1)
        # Check heralds are in correct locations
        assert 0 in circuit._external_heralds.input
        assert 2 in circuit._external_heralds.output
        assert 3 in circuit._external_heralds.input
        assert 5 in circuit._external_heralds.output

    def test_heralded_circuit_addition_circ_modification(self):
        """
        Checks that components in an existing circuit are modified correctly
        when a heralded sub-circuit is added.
        """
        # Place beam splitters across 0 & 1 and 2 & 3
        circuit = PhotonicCircuit(4)
        circuit.bs(0)
        circuit.bs(2)
        # Create heralded sub-circuit and add
        sub_circ = PhotonicCircuit(4)
        sub_circ.herald((0, 0), 1)
        sub_circ.herald((3, 3), 1)
        circuit.add(sub_circ, 1)
        # Confirm beam splitters now act on modes 0 & 2 and 3 & 5
        spec = circuit._get_circuit_spec()
        # Get relevant elements from spec
        assert spec[0].mode_1 == 0
        assert spec[0].mode_2 == 2
        assert spec[1].mode_1 == 3
        assert spec[1].mode_2 == 5

    def test_heralded_circuit_addition_circ_modification_2(self):
        """
        Checks that components in an existing circuit are modified correctly
        when a heralded sub-circuit is added in a different location.
        """
        # Place beam splitters across 0 & 1 and 2 & 3
        circuit = PhotonicCircuit(4)
        circuit.bs(0)
        circuit.bs(2)
        # Create heralded sub-circuit and add
        sub_circ = PhotonicCircuit(4)
        sub_circ.herald((0, 0), 1)
        sub_circ.herald((3, 3), 1)
        circuit.add(sub_circ, 0)
        # Confirm beam splitters now act on modes 0 & 2 and 3 & 5
        spec = circuit._get_circuit_spec()
        # Get relevant elements from spec
        assert spec[0].mode_1 == 1
        assert spec[0].mode_2 == 2
        assert spec[1].mode_1 == 4
        assert spec[1].mode_2 == 5

    def test_heralded_circuit_addition_circ_modification_3(self):
        """
        Checks that components in an existing circuit are modified correctly
        when a heralded sub-circuit is added with a different configuration.
        """
        # Place beam splitters across 0 & 1 and 2 & 3
        circuit = PhotonicCircuit(4)
        circuit.bs(0)
        circuit.bs(2)
        # Create heralded sub-circuit and add
        sub_circ = PhotonicCircuit(4)
        sub_circ.herald((0, 1), 1)
        sub_circ.herald((1, 0), 1)
        circuit.add(sub_circ, 1)
        # Confirm beam splitters now act on modes 0 & 2 and 3 & 5
        spec = circuit._get_circuit_spec()
        # Get relevant elements from spec
        assert spec[0].mode_1 == 0
        assert spec[0].mode_2 == 3
        assert spec[1].mode_1 == 4
        assert spec[1].mode_2 == 5

    def test_heralded_circuit_addition_circ_modification_all_herald(self):
        """
        Checks that components in an existing circuit are modified correctly
        when a heralded sub-circuit is added with all modes heralded.
        """
        # Place beam splitters across 0 & 1 and 2 & 3
        circuit = PhotonicCircuit(4)
        circuit.bs(0)
        circuit.bs(2)
        # Create heralded sub-circuit and add
        sub_circ = PhotonicCircuit(4)
        sub_circ.herald((0, 0), 1)
        sub_circ.herald((1, 1), 1)
        sub_circ.herald((2, 2), 1)
        sub_circ.herald((3, 3), 1)
        circuit.add(sub_circ, 1)
        # Confirm beam splitters now act on modes 0 & 2 and 3 & 5
        spec = circuit._get_circuit_spec()
        # Get relevant elements from spec
        assert spec[0].mode_1 == 0
        assert spec[0].mode_2 == 5
        assert spec[1].mode_1 == 6
        assert spec[1].mode_2 == 7

    def test_heralded_two_circuit_addition_circ_modification(self):
        """
        Checks that components in an existing circuit are modified correctly
        when two heralded sub-circuits are added.
        """
        # Place beam splitters across 0 & 1 and 2 & 3
        circuit = PhotonicCircuit(4)
        circuit.bs(0)
        circuit.bs(2)
        # Create heralded sub-circuit and add
        sub_circ = PhotonicCircuit(4)
        sub_circ.herald((0, 1), 1)
        sub_circ.herald((2, 0), 1)
        circuit.add(sub_circ, 2)
        circuit.add(sub_circ, 1)
        # Confirm beam splitters now act on modes 0 & 2 and 3 & 5
        spec = circuit._get_circuit_spec()
        # Get relevant elements from spec
        assert spec[0].mode_1 == 0
        assert spec[0].mode_2 == 2
        assert spec[1].mode_1 == 5
        assert spec[1].mode_2 == 7

    def test_herald_varied_on_sub_circuit(self):
        """
        Checks that heralds are correctly replicated when the input and output
        of a herald are on different modes.
        """
        circ = PhotonicCircuit(4)
        sub_circ = PhotonicCircuit(3)
        sub_circ.herald((0, 2), (1, 2))
        circ.add(sub_circ)
        assert circ.heralds.input[0] == 1
        assert circ.heralds.output[0] == 2

    def test_input_modes(self):
        """
        Checks input modes attribute returns n_modes value when no heralds are
        used.
        """
        circuit = PhotonicCircuit(randint(4, 10))
        assert circuit.input_modes == circuit.n_modes

    def test_input_modes_heralds(self):
        """
        Checks input modes attribute returns less then n_modes value when
        heralds are included.
        """
        n = randint(6, 10)
        circuit = PhotonicCircuit(n)
        circuit.herald((1, 4), 1)
        circuit.herald((3, 3), 2)
        assert circuit.input_modes == n - 2

    def test_input_modes_heralds_sub_circuit(self):
        """
        Checks input modes attribute returns original n_modes value when
        heralds are included as part of a sub-circuit and then added to the
        larger circuit.
        """
        n = randint(9, 10)
        circuit = PhotonicCircuit(n)
        sub_circuit = PhotonicCircuit(4)
        sub_circuit.herald((0, 0), 1)
        sub_circuit.herald((3, 1), 0)
        circuit.add(sub_circuit, 2)
        assert circuit.input_modes == n

    def test_input_modes_heralds_sub_circuit_original_heralds(self):
        """
        Checks input modes attribute returns original n_modes value when
        heralds are included as part of a sub-circuit and then added to the
        larger circuit, with the larger circuit also containing heralds.
        """
        n = randint(9, 10)
        circuit = PhotonicCircuit(n)
        circuit.herald((1, 4), 1)
        circuit.herald((3, 3), 2)
        sub_circuit = PhotonicCircuit(4)
        sub_circuit.herald((0, 0), 1)
        sub_circuit.herald((3, 1), 0)
        circuit.add(sub_circuit, 2)
        assert circuit.input_modes == n - 2

    @pytest.mark.parametrize(
        ("initial_modes", "final_modes"),
        [((0,), (0, 2)), ((2,), (5, 6)), ((2, 3), (5, 6)), ((0, 2), (0, 5))],
    )
    def test_bs_correct_modes(self, initial_modes, final_modes):
        """
        Checks that when a circuit has internal modes then the beam splitter is
        applied correctly across the other modes.
        """
        circuit = PhotonicCircuit(7)
        circuit._PhotonicCircuit__internal_modes = [1, 3, 4]
        # Add bs on modes
        circuit.bs(*initial_modes)
        # Then check modes are converted to correct values
        assert (
            circuit._PhotonicCircuit__circuit_spec[0].mode_1 == final_modes[0]
        )
        assert (
            circuit._PhotonicCircuit__circuit_spec[0].mode_2 == final_modes[1]
        )

    @pytest.mark.parametrize("convention", ["Rx", "H"])
    def test_bs_valid_convention(self, convention):
        """
        Checks beam splitter accepts valid conventions.
        """
        circuit = PhotonicCircuit(3)
        circuit.bs(0, convention=convention)

    @pytest.mark.parametrize("convention", ["Rx", "H"])
    def test_bs_valid_convention_splitting_ratio(self, convention):
        """
        Checks all beam splitter conventions with reflectivity of 0.5 implement
        the intended splitting ratio.
        """
        circuit = PhotonicCircuit(3)
        circuit.bs(0, convention=convention)
        assert abs(circuit.U[0, 0]) ** 2 == pytest.approx(0.5)

    def test_bs_invalid_convention(self):
        """
        Checks a ValueError is raised if an invalid beam splitter convention is
        set in bs.
        """
        circuit = PhotonicCircuit(3)
        with pytest.raises(ValueError):
            circuit.bs(0, convention="Not valid")

    @pytest.mark.parametrize("value", [-0.5, 1.4, "0.5", True])
    def test_bs_invalid_reflectivity(self, value):
        """
        Checks a ValueError or TypeError is raised if an invalid beam splitter
        reflectivity is set.
        """
        circuit = PhotonicCircuit(3)
        with pytest.raises((ValueError, TypeError)):
            circuit.bs(0, reflectivity=value)

    @pytest.mark.parametrize("value", [0.1, 1, Parameter(1)])
    def test_ps_valid_phi(self, value):
        """
        Checks valid phase shift values are accepted by the phase shifter
        component.
        """
        circuit = PhotonicCircuit(3)
        circuit.ps(1, value)

    @pytest.mark.parametrize("value", [True, None, "test", "1"])
    def test_ps_invalid_phi(self, value):
        """
        Checks invalid phase shift values raise an exception.
        """
        circuit = PhotonicCircuit(3)
        with pytest.raises((ValueError, TypeError)):
            circuit.ps(1, value)

    def test_mode_swaps_invalid(self):
        """
        Checks a ValueError is raised if an invalid mode swap configuration is
        set.
        """
        circuit = PhotonicCircuit(3)
        with pytest.raises(ValueError):
            circuit.mode_swaps({0: 1})

    def test_get_all_params(self):
        """
        Tests get_all_params method correctly identifies all parameters added
        within a circuit.
        """
        p1 = Parameter(0.5)
        p2 = Parameter(2)
        p3 = Parameter(convert.db_loss_to_decimal(3))
        # Create circuit with parameters
        circuit = PhotonicCircuit(4)
        circuit.bs(0, reflectivity=p1)
        circuit.ps(2, p2)
        circuit.loss(3, p3)
        # Recover all params and then check
        all_params = circuit.get_all_params()
        for param in [p1, p2, p3]:
            assert param in all_params

    def test_get_all_params_grouped_circ(self):
        """
        Tests get_all_params method correctly identifies all parameters added
        within a grouped circuit.
        """
        p1 = Parameter(0.5)
        p2 = Parameter(2)
        p3 = Parameter(convert.db_loss_to_decimal(3))
        # Create sub-circuit with parameters
        sub_circuit = PhotonicCircuit(4)
        sub_circuit.bs(0, reflectivity=p1)
        sub_circuit.ps(2, p2)
        sub_circuit.loss(3, p3)
        # Then add to larger circuit
        circuit = PhotonicCircuit(5)
        circuit.add(sub_circuit, 1, group=True)
        # Recover all params and then check
        all_params = circuit.get_all_params()
        for param in [p1, p2, p3]:
            assert param in all_params

    def test_edit_circuit_spec(self):
        """
        Checks an exception is raised if a circuit spec is modified with an
        invalid value.
        """
        circuit = PhotonicCircuit(4)
        circuit._PhotonicCircuit__circuit_spec = [["test", None]]
        with pytest.raises(CircuitCompilationError):
            circuit._build()

    def test_all_circuit_spec_components(self):
        """
        Confirms that all elements in a created circuit spec are components
        """
        circuit = PhotonicCircuit(6)
        circuit.bs(0, 1)
        circuit.ps(2, 2)
        circuit.mode_swaps({0: 1, 1: 2, 2: 0})
        circuit.barrier()
        circuit.loss(3, 1)
        # Define sub-circuit to add
        sub_circuit = PhotonicCircuit(4)
        circuit.bs(0, 1)
        circuit.ps(2, 2)
        circuit.mode_swaps({0: 1, 1: 2, 2: 0})
        circuit.barrier()
        circuit.loss(3, 1)
        # Also include CNOT
        circuit.add(sub_circuit, 1, group=True)
        circuit.add(CNOT(), 1)
        # Check all are components
        assert all(
            isinstance(c, Component) for c in circuit._get_circuit_spec()
        )
        # Unpack the circuit spec and repeat
        circuit.unpack_groups()
        assert all(
            isinstance(c, Component) for c in circuit._get_circuit_spec()
        )

    @pytest.mark.parametrize(
        "path",
        ["test", "test.svg", "test.png", "tests/test.svg", "tests\\test.svg"],
    )
    def test_circuit_saving_svg(self, path):
        """
        Checks that a figure can be created for a range target paths.
        """
        circuit = PhotonicCircuit(6)
        circuit.add(CNOT())
        circuit.bs(0)
        circuit.ps(0, Parameter(3, label="μ"))
        circuit.loss(2, 0.5)
        circuit.mode_swaps({0: 2, 2: 0})
        circuit.barrier()
        # Save circuit
        circuit.save_figure(path, svg=True)
        # Check it exists
        path = Path(path).with_suffix(".svg")
        assert path.exists()
        path.unlink()

    def test_circuit_saving_png(self):
        """
        Checks that a png figure can either be saved or the appropriate error is
        raised depending on the installed dependencies.
        """
        path = "test.png"
        circuit = PhotonicCircuit(6)
        circuit.add(CNOT())
        circuit.bs(0)
        circuit.ps(0, Parameter(3, label="μ"))
        circuit.loss(2, 0.5)
        circuit.mode_swaps({0: 2, 2: 0})
        circuit.barrier()
        # Save circuit
        try:
            circuit.save_figure(path, svg=True)
        except LightworksDependencyError:
            pass
        else:
            # Check it exists
            path = Path(path).with_suffix(".svg")
            assert path.exists()
            path.unlink()

    def test_circuit_equivalence(self):
        """
        Checks that the equals comparison between two circuits returns true when
        they are the same.
        """
        # Define two circuits, changing the order in which components are added
        c1 = PhotonicCircuit(5)
        c1.bs(0, reflectivity=0.4)
        c1.bs(3, reflectivity=0.55)
        c1.ps(2, 0.531, loss=0.3)
        c2 = PhotonicCircuit(5)
        c2.ps(2, 0.531, loss=0.3)
        c2.bs(3, reflectivity=0.55)
        c2.bs(0, reflectivity=0.4)
        # Check equivalence
        assert c1 == c2

    def test_circuit_non_equivalence(self):
        """
        Checks that the equals comparison between two circuits returns false
        when the unitary differs.
        """
        # Define two circuits, changing the order in which components are added
        c1 = PhotonicCircuit(5)
        c1.bs(0, reflectivity=0.4)
        c1.bs(3, reflectivity=0.55)
        c1.ps(2, 0.531, loss=0.3)
        c2 = PhotonicCircuit(5)
        c2.ps(2, 0.2, loss=0.3)
        c2.bs(3, reflectivity=0.3)
        c2.bs(0, reflectivity=0.45)
        # Check equivalence
        assert c1 != c2

    def test_circuit_non_equivalence_modes(self):
        """
        Checks that the equals comparison between two circuits returns false
        when the number of modes is different.
        """
        # Define two circuits, changing the order in which components are added
        c1 = PhotonicCircuit(5)
        c2 = PhotonicCircuit(2)
        # Check equivalence
        assert c1 != c2

    def test_circuit_non_equivalence_loss_modes(self):
        """
        Checks that the equals comparison between two circuits returns false
        when the number of loss modes is different.
        """
        # Define two circuits, changing the order in which components are added
        c1 = PhotonicCircuit(5)
        c2 = PhotonicCircuit(5)
        c2.loss(0, 0.5)
        # Check equivalence
        assert c1 != c2


class TestUnitary:
    """
    Unit tests to confirm correct functioning of the Unitary class when various
    operations are performed.
    """

    def test_unitary_assignment(self):
        """Checks that a unitary is correctly assigned with the component."""
        u = random_unitary(4)
        unitary = Unitary(u)
        assert (u == unitary.U).all()

    def test_non_unitary_assignment(self):
        """
        Checks that errors are raised when non-unitary matrices are assigned.
        """
        # Non-square unitary
        u = random_unitary(4)
        u = u[:, :-2]
        with pytest.raises(ValueError):
            Unitary(u)
        # Non-unitary matrix
        u2 = random_unitary(4)
        u2[0, 0] = 1
        with pytest.raises(ValueError):
            Unitary(u2)

    def test_circuit_addition_to_unitary(self):
        """
        Confirm that addition of a circuit to the Unitary object works as
        expected.
        """
        u = Unitary(random_unitary(6, seed=95))
        circ = PhotonicCircuit(4)
        circ.bs(0)
        circ.bs(2, loss=convert.db_loss_to_decimal(0.5))
        circ.bs(1, loss=convert.db_loss_to_decimal(0.2))
        u.add(circ, 1)
        assert u.U[0, 0] == pytest.approx(
            -0.27084817086493 - 0.176576418865914j, 1e-8
        )
        assert u.U[1, 2] == pytest.approx(
            0.232353190742325 - 0.444902420616067j, 1e-8
        )
        assert u.U[4, 3] == pytest.approx(
            -0.31290267006132 - 0.091957924939349j, 1e-8
        )

    def test_unitary_is_circuit_child(self):
        """
        Checks that the unitary object is a child class of the PhotonicCircuit
        object.
        """
        u = Unitary(random_unitary(4))
        assert isinstance(u, PhotonicCircuit)

    def test_n_mode_retrival(self):
        """
        Confirms n_mode attribute retrieval works for unitary component.
        """
        u = Unitary(random_unitary(4))
        assert u.n_modes == 4

    def test_unitary_parameters(self):
        """
        Checks parameters can be used to set and update a unitary.
        """
        th = sp.Symbol("theta")
        unitary_mat = sp.Matrix(
            [
                [sp.cos(th / 2), -1j * sp.sin(th / 2)],
                [-1j * sp.sin(th / 2), sp.cos(th / 2)],
            ]
        )
        theta = Parameter(0.1)
        # Create component
        unitary = ParameterizedUnitary(unitary_mat, {"theta": theta})
        component = Unitary(unitary)
        # Get unitary
        u1 = component.U
        # Update parameter and get new unitary
        theta.set(0.6)
        u2 = component.U
        # Check not equivalent
        assert (u1 != u2).all()

    def test_invalid_unitary(self):
        """
        Checks an error is raised when an initially valid unitary is updated via
        parameters to be invalid.
        """
        th = sp.Symbol("theta")
        unitary_mat = sp.Matrix([[th, 0], [0, th]])
        theta = Parameter(1)
        # Create component
        unitary = ParameterizedUnitary(unitary_mat, {"theta": theta})
        component = Unitary(unitary)
        theta.set(2)
        with pytest.raises(CircuitCompilationError):
            unitary = component.U



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from types import FunctionType, MethodType
from typing import Any

import matplotlib.figure
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from lightworks.sdk.state import State
from lightworks.sdk.utils import PostSelectionFunction
from lightworks.sdk.utils.exceptions import ResultCreationError
from lightworks.sdk.utils.post_selection import PostSelectionType

from .result import Result


class SamplingResult(Result[State, int]):
    """
    Stores results data from a sampling experiment in the emulator. There is
    then a range of options for displaying the data, or alternatively the data
    can be accessed directly using the [] operator on the class to select which
    output is required.

    Args:

        results (dict | np.ndarray) : The results which are to be stored.

        input (State) : The input state used in the sampling experiment.

    """

    def __init__(
        self, results: dict[State, int], input: State, **kwargs: Any
    ) -> None:
        super().__init__(results)
        if not isinstance(input, State):
            raise ResultCreationError("Input state should have type State.")
        self.__input = input
        self.__outputs = list(results.keys())
        # Store any additional provided data from kwargs as attributes
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def input(self) -> State:
        """The input state used in the sampling experiment."""
        return self.__input

    @property
    def outputs(self) -> list[State]:
        """All outputs measured in the sampling experiment."""
        return self.__outputs

    def __getitem__(self, item: State) -> int:
        """Custom get item behaviour - used when object accessed with []."""
        if not isinstance(item, State):
            raise TypeError("Get item value must be a State.")
        if item not in self:
            raise KeyError("Provided output state not in data.")
        return super().__getitem__(item)

    def map(
        self,
        mapping: Callable[[State, Any], State],
        *args: Any,
        **kwargs: Any,
    ) -> "SamplingResult":
        """
        Performs a generic remapping of states based on a provided function.
        """
        if not isinstance(mapping, FunctionType | MethodType):
            raise TypeError(
                "Provided mapping should be a callable function which accepts "
                "and returns a State object."
            )
        mapped_result: dict[State, int] = {}
        for out_state, val in self.items():
            new_s = mapping(out_state, *args, **kwargs)
            if new_s in mapped_result:
                mapped_result[new_s] += val
            else:
                mapped_result[new_s] = val
        return self._recombine_mapped_result(mapped_result)

    def apply_post_selection(
        self, post_selection: PostSelectionType | Callable[[State], bool]
    ) -> "SamplingResult":
        """
        Applies an additional post-selection criteria to the stored result and
        returns this as a new object.
        """
        if isinstance(post_selection, FunctionType | MethodType):
            post_selection = PostSelectionFunction(post_selection)
        if not isinstance(post_selection, PostSelectionType):
            raise TypeError(
                "Provided post_selection should either be a PostSelection "
                "object or a callable function which accepts a state and "
                "returns a boolean to indicate whether the state is valid."
            )
        return self._recombine_mapped_result(
            {
                state: val
                for state, val in self.items()
                if post_selection.validate(state)
            }
        )

    def _recombine_mapped_result(
        self, mapped_result: dict[State, int]
    ) -> "SamplingResult":
        """Creates a new Result object from mapped data."""
        return SamplingResult(mapped_result, self.input)

    def plot(
        self,
        show: bool = True,
        state_labels: dict[State, str | State] | None = None,
    ) -> tuple[matplotlib.figure.Figure, plt.Axes] | None:
        """
        Create a plot of the data contained in the result. This will either
        take the form of a heatmap or bar chart, depending on the nature of the
        data contained in the Result object.

        Args:

            show (bool, optional) : Can be used to automatically show the
                created plot with show instead of returning the figure and
                axes.

            state_labels (dict, optional) : Provided a dictionary which can be
                used to specify alternate labels for each of the states when
                plotting. The keys of this dictionary should be States and the
                values should be strings or States.

        """
        if state_labels is None:
            state_labels = {}
        # Check provided state labels are valid
        for state, label in state_labels.items():
            if not isinstance(state, State):
                raise TypeError("Keys of state_labels dict should be States.")
            # Convert values from state_labels to strings if not already
            state_labels[state] = str(label)

        fig, ax = plt.subplots(figsize=(7, 6))
        x_data = range(len(self))
        ax.bar(x_data, list(self.values()))
        ax.set_xticks(x_data)
        labels = [
            state_labels[s] if s in state_labels else str(s) for s in self
        ]
        ax.set_xticklabels(labels, rotation=90)  # type: ignore[arg-type]
        ax.set_xlabel("State")
        ax.set_ylabel("Counts")

        # Optionally use show on plot if specified
        if show:
            plt.show()
            return None
        return (fig, ax)

    def print_outputs(self) -> None:
        """
        Print the output results for each input into the system. This is
        compatible with all possible result types.
        """
        to_print = str(self.input) + " -> "
        for ostate, p in self.items():
            to_print += str(ostate) + " : " + str(p) + ", "
        to_print = to_print[:-2]
        print(to_print)  # noqa: T201

    def display_as_dataframe(self, threshold: float = 1e-12) -> pd.DataFrame:
        """
        Function to display the results of a given simulation in a dataframe
        format. Either the probability amplitudes of the state, or the actual
        probabilities can be displayed.

        Args:

            threshold (float, optional) : Threshold to control at which point
                value are rounded to zero. If looking for very small amplitudes
                this may need to be lowered.

            conv_to_probability (bool, optional) : In the case that the result
                is a probability amplitude, setting this to True will convert
                it into a probability. If it is not a probability amplitude
                then this setting will have no effect.

        Returns:

            pd.Dataframe : A dataframe with the results and input and output
                states as labels.

        """
        # Convert state vectors into strings
        in_strings = [str(self.input)]
        out_strings = [str(s) for s in self.outputs]
        # Switch to probability if required
        data = np.array(list(self.values()))
        # Apply thresholding to values
        for i in range(data.shape[0]):
            val = data[i]
            re = np.real(val) if abs(np.real(val)) > threshold else 0
            im = np.imag(val) if abs(np.imag(val)) > threshold else 0
            data[i] = re if abs(im) == 0 else re + 1j * im
        # Convert array to floats when not non complex results used
        data = data.astype(int)
        # Create dataframe
        results = pd.DataFrame(data, index=out_strings, columns=in_strings)
        return results.transpose()



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines all custom exceptions used as part of the SDK section of lightworks.
In general, the approach is to use the built-in exceptions where possible, but
provide custom exceptions where the extra context would be useful, or it may
cause confusion as to the root of an issue if a generic type is used.
"""


class LightworksError(Exception):
    """
    Generic error from which all other errors are derived.
    """


class StateError(LightworksError):
    """
    Error relating to issues with a provided State
    """


class ModeRangeError(LightworksError):
    """
    Error for specific errors arising when a provided mode is outside of the
    circuit range.
    """


class CircuitCompilationError(LightworksError):
    """
    For all errors that arise during compilation of a circuit.
    """


class DisplayError(LightworksError):
    """
    Used when specific errors during the Display methods.
    """


class ParameterValueError(LightworksError):
    """
    For errors in the setting of a Parameter value.
    """


class ParameterBoundsError(LightworksError):
    """
    For errors in the setting of Parameter bounds.
    """


class ParameterDictError(LightworksError):
    """
    Exceptions relating to ParameterDict behaviour.
    """


class DecompositionUnsuccessful(LightworksError):  # noqa: N818
    """
    For situations in which a unitary decomposition procedure is unsuccessful.
    """


class ResultError(LightworksError):
    """
    For specific errors which occur when using a Result object.
    """


class ResultCreationError(ResultError):
    """
    For specific errors which occur when creating a Result object.
    """


class ProbabilityDistributionError(LightworksError):
    """
    For errors occurring with the ProbabilityDistribution object.
    """


class BackendError(LightworksError):
    """
    Raised when errors occur in the Backend object.
    """


class ModeMismatchError(LightworksError):
    """
    For use in simulation objects when there is a mode mismatch between
    provided states/circuit.
    """


class PhotonNumberError(LightworksError):
    """
    For use in simulation objects when there is a photon number mismatch
    between inputs and/or outputs.
    """


class SamplerError(LightworksError):
    """
    Specific error in the Sampler objects.
    """


class LightworksDependencyError(LightworksError):
    """
    For errors relating to a dependency of lightworks.
    """



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Models a non-ideal source of single photons, it can be used to build input
state variations, accounting for a non-ideal brightness, impurity in the single
photon stream and indistinguishability between different photons.
"""

from numbers import Number

from lightworks.emulator.state import AnnotatedState
from lightworks.sdk.state import State


class Source:
    """
    Simulates an imperfect single photon source to model properties of photon
    inputs in the Sampler.

    Args:

        purity (float) : The purity of the photon stream from the single photon
            source, should be given as a decimal. Note that it is assumed that
            any additional photons are distinguishable from the source photons.

        brightness (float) : The brightness of the single photon source as a
            decimal value.

        indistinguishability (float) : The indistinguishability of the input
            photons.

        probability_threshold (float, optional) : Set a probability threshold
            for the input, this can be used to remove any inputs with
            probabilities below this value. Note, this should be used with some
            caution as it may reduce the overall accuracy of the simulation.

    """

    def __init__(
        self,
        purity: float = 1,
        brightness: float = 1,
        indistinguishability: float = 1,
        probability_threshold: float = 1e-9,
    ) -> None:
        self.purity = purity
        self.brightness = brightness
        self.indistinguishability = indistinguishability
        self.probability_threshold = probability_threshold

    def __str__(self) -> str:
        return (
            f"Source(purity = {self.purity}, brightness = {self.brightness}, "
            f"indistinguishability = {self.indistinguishability}, "
            f"probability threshold = {self.probability_threshold})"
        )

    def __repr__(self) -> str:
        return "lightworks.emulator." + str(self)

    @property
    def brightness(self) -> float:
        """The output efficiency of the single photon source."""
        return self.__brightness

    @brightness.setter
    def brightness(self, value: float) -> None:
        quantity_check(value, "brightness")
        self.__brightness = value

    @property
    def purity(self) -> float:
        """The purity as a source, given as a decimal. Purity = 1 - g2."""
        return self.__purity

    @purity.setter
    def purity(self, value: float) -> None:
        # Special check for purity as it requires value >= 0.5
        if not 0.5 < value <= 1:
            raise ValueError("Value of purity should be in range (0.5,1].")
        quantity_check(value, "purity")
        self.__purity = value

    @property
    def indistinguishability(self) -> float:
        """The measured indistinguishability of the single photon source."""
        return self.__indistinguishability

    @indistinguishability.setter
    def indistinguishability(self, value: float) -> None:
        quantity_check(value, "indistinguishability")
        self.__indistinguishability = value

    @property
    def probability_threshold(self) -> float:
        """
        Define a probability threshold below which single input states are
        discarded.
        """
        return self.__probability_threshold

    @probability_threshold.setter
    def probability_threshold(self, value: float) -> None:
        quantity_check(value, "probability_threshold")
        self.__probability_threshold = value

    def check_number(self, state: State) -> int:
        """
        Check the number of possible inputs states that a given target input
        and source properties will produce.

        Args:

            state (State) : The target input state produced by the source.

        Returns:

            int : The number of states contained in the generated source input
                state probability distribution.

        """
        return len(self._build_statistics(state))

    def _build_statistics(
        self, state: State
    ) -> dict[State, float] | dict[AnnotatedState, float]:
        """
        Determine the true input statistics for a given target input and the
        provided properties of the single photon source.

        Args:

            state (State) : The target input state for the calculation.

        Returns:

            dict : The created statistics dictionary containing all possible
                inputs and their probabilities. This returned inputs will
                either be States or AnnotatedStates depending on the required
                input properties.

        """
        # Basic method
        if self.purity == 1 and self.indistinguishability == 1:
            basic_stats_dict = self._build_statistics_basic(state)
            return apply_threshold(basic_stats_dict, self.probability_threshold)
        # Full method
        stats_dict = self._build_statistics_full(state)
        return apply_threshold(stats_dict, self.probability_threshold)

    def _build_statistics_basic(self, state: State) -> dict[State, float]:
        """
        Build a basic version of statistics which only accounts for source
        brightness.

        Args:

            state (State) : The target input state for the calculation.

        Returns:

            dict : The created statistics dictionary containing all possible
                inputs and their probabilities.

        """
        n_modes = len(state)
        stats: list[tuple[float, list[int]]] = []
        # Loop over each mode
        for mode, count in enumerate(state):
            if not count:
                continue
            for _i in range(count):
                # For a given photon in a mode work out all possible input
                # combinations
                sub_s = [
                    (
                        self.brightness,
                        [1 if mode == j else 0 for j in range(n_modes)],
                    )
                ]
                # Add zero mode from less than perfect brightness
                if self.brightness < 1:
                    sub_s.append((1 - self.brightness, [0] * n_modes))
                # Create stats list on first run
                if not stats:
                    stats = sub_s
                # Then combine stats with new calculated states to find new
                # probabilities
                else:
                    new_stats = []
                    for p1, s1 in stats:
                        for p2, s2 in sub_s:
                            new_stats.append(
                                (
                                    p1 * p2,
                                    [
                                        x + y
                                        for x, y in zip(s1, s2, strict=True)
                                    ],
                                )
                            )
                    stats = new_stats
        # Combine any duplicate results
        stats_dict = {}
        for probs, s in stats:
            s_conv = State(s)
            if s_conv not in stats_dict:
                stats_dict[s_conv] = probs
            else:
                stats_dict[s_conv] += probs
        # Catch any empty returns and give input
        if not stats_dict:
            stats_dict[state] = 1.0
        return stats_dict

    def _build_statistics_full(
        self, state: State
    ) -> dict[AnnotatedState, float]:
        """
        Builds a full version of statistics which can account for the
        brightness, indistinguishability and purity of the source.

        Args:

            state (State) : The target input state for the calculation.

        Returns:

            dict : The created statistics dictionary containing all possible
                inputs and their probabilities. All inputs will be represented
                as annotated states, which use labelling to denote photons
                which are distinguishable from each other.

        """
        # Reset counter
        self._counter = 1
        # Get the full distribution and then perform remapping
        stats_dict = self._full_distribution(state)
        return self._remap_distribution(stats_dict)

    def _remap_distribution(
        self, distribution: dict[AnnotatedState, float]
    ) -> dict[AnnotatedState, float]:
        """
        Remaps a provided distribution to a common form, enabling equivalent
        entries to be removed.
        """
        # Remap states to be correctly ordered
        new_dist = {}
        for state, p in distribution.items():
            # Determine all labels in a given state and create a mapping which
            # converts all labels to their minimum value
            all_labels = []
            for mode in state:
                all_labels += mode
            all_labels = list(dict.fromkeys(all_labels))
            mapping = {all_labels[i]: i for i in range(len(all_labels))}
            # Apply determined mapping to each state
            list_state = state.s
            for j, mode in enumerate(list_state):
                for i, m in enumerate(mode):
                    mode[i] = mapping[m]
                list_state[j] = mode
            state = AnnotatedState(list_state)  # noqa: PLW2901
            # Add new state to new distribution
            if state not in new_dist:
                new_dist[state] = p
            else:
                new_dist[state] += p

        return new_dist

    def _full_distribution(self, state: State) -> dict[AnnotatedState, float]:
        """
        Calculates the full input state probability distribution for a given
        state.
        """
        # Find groups of empty modes
        to_group, to_skip = group_empty_modes(state)
        input_dist: dict[AnnotatedState, float] = {}
        # Loop over each mode and find distribution
        for i, n in enumerate(state):
            if i in to_skip:
                continue
            # Added groups of empty photons
            if i in to_group:
                if not input_dist:
                    input_dist = {AnnotatedState([[]] * len(to_group[i])): 1.0}
                else:
                    input_dist = {
                        s1 + AnnotatedState([[]] * len(to_group[i])): p1
                        for s1, p1 in input_dist.items()
                    }
            else:
                calc_dist = self._single_mode_distribution(n)
                # Then combine with the full input distribution
                if not input_dist:
                    input_dist = calc_dist
                else:
                    new_dist = {}
                    for s1, p1 in input_dist.items():
                        for s2, p2 in calc_dist.items():
                            new_dist[s1 + s2] = p1 * p2
                    input_dist = new_dist

        return input_dist

    def _single_mode_distribution(
        self, n_photons: int
    ) -> dict[AnnotatedState, float]:
        """
        Finds the input state statistics for a number of photons on a given
        mode.
        """
        # Check for special case in which a mode is empty
        if n_photons == 0:
            return {AnnotatedState([[]]): 1.0}
        mode_dist: list[tuple[list[int], float]] = []
        # Get distribution for each photon and combine with mode distribution
        for _i in range(n_photons):
            calc_dist = self._single_photon_distribution()
            if not mode_dist:
                mode_dist = calc_dist
            else:
                new_dist = [
                    (d1[0] + d2[0], d1[1] * d2[1])
                    for d1 in mode_dist
                    for d2 in calc_dist
                ]
                mode_dist = new_dist
        # Sort labels for easier processing
        for d in mode_dist:
            d[0].sort()
        # Convert to annotated states and add to a dictionary
        dist_dict = {}
        for s, p in mode_dist:
            state = AnnotatedState([s])
            if state not in dist_dict:
                dist_dict[state] = p
            else:
                dist_dict[state] += p
        return dist_dict

    def _single_photon_distribution(self) -> list[tuple[list[int], float]]:
        """
        Calculates the distribution for a single photon with imperfect
        properties. This uses the probabilities as shown on page 15 of
        arXiv:2211.15626v1. It is assumed that the distinguishable and
        multi-photon components are independent between different single photon
        distributions.
        """
        # Define some useful quantities
        nu = self.brightness
        p_i = self.indistinguishability**0.5
        p_d = 1 - p_i
        # Find pure state probability using purity
        p1 = purity_to_prob(self.purity)
        p2 = 1 - p1
        # Define all required probabilities
        c0 = 1 - nu * (p1 + p2 * nu + 2 * (1 - nu) * p2)
        c1 = p_i * nu * (p1 + (1 - nu) * p2)
        c1d = p_d * nu * (p1 + (1 - nu) * p2)
        c1dp = nu * (1 - nu) * p2
        c12d = nu**2 * p_i * p2
        c1d2d = nu**2 * p_d * p2
        # Determine unique labels for distinguishable and multi-photon photons
        dpc = self._counter
        mpc = self._counter + 1
        self._counter += 2
        # Collect all possible supported entries
        to_add = [
            ([], c0),
            ([0], c1),
            ([dpc], c1d),
            ([mpc], c1dp),
            ([0, mpc], c12d),
            ([dpc, mpc], c1d2d),
        ]
        return [(s, p) for s, p in to_add if p > 0]


def apply_threshold(
    distribution: dict[State, float] | dict[AnnotatedState, float],
    threshold: float,
) -> dict[State, float] | dict[AnnotatedState, float]:
    """
    Applies a threshold to a provided source probability distribution
    """
    thresholded_dict = {s: p for s, p in distribution.items() if p >= threshold}
    # Re-normalize after removing values
    total = sum(thresholded_dict.values())
    for s, p in thresholded_dict.items():
        thresholded_dict[s] = p / total
    return thresholded_dict  # type: ignore[return-value]


def group_empty_modes(state: State) -> tuple[dict[int, list[int]], list[int]]:
    """
    Used to find groups of empty modes, this can be used to speed up the source
    distribution calculation.
    """
    to_group = {}
    to_skip = []
    # Loop over each mode in the state
    for i, s in enumerate(state):
        # Skip any modes already grouped or the last mode
        if i in to_skip or i == len(state) - 1:
            continue
        if s == 0:
            # If only a single empty mode then skip
            if state[i + 1] > 0:
                continue
            # Find how many empty modes there are
            n = 1
            while state[i + n] == 0:
                n += 1
                if i + n > len(state) - 1:
                    break
            to_group[i] = list(range(i, i + n))
            to_skip += list(range(i + 1, i + n))

    return (to_group, to_skip)


def purity_to_prob(purity: float) -> float:
    """
    Return single photon probability for a given purity, probability of a two
    photon state will be twice this.
    """
    if purity <= 0.5:
        raise ValueError("Purity values below 0.5 not supported.")
    if purity < 1:
        g2 = 1 - purity
        b = 2 * (1 - (1 / g2))
        return 1 - (-b - (b**2 - 4) ** 0.5) / 2
    return 1


def quantity_check(value: float, name: str) -> None:
    """
    General function to confirm that a quantity is numerical and lies in the
    range [0,1].

    Args:

        value (float) : The quantity that is to be checked.

        name (str) : A name to use for the quantity when displaying error
            messages.

    Raises:

        TypeError : When quantity isn't numeric.

        ValueError : When quantity is not within the range [0,1].

    """
    if not isinstance(value, Number) or isinstance(value, bool):
        msg = f"{name} should be a numerical value."
        raise TypeError(msg)
    if not 0 <= value <= 1:
        msg = f"Value of {name} should be in range [0,1]."
        raise ValueError(msg)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Script to store various useful functions for the simulation aspect of the code.
"""

from collections.abc import Iterable


def fock_basis(n_modes: int, n_photons: int) -> list[list[int]]:
    """Returns the Fock basis for n photons in N modes."""
    return list(_sums(n_modes, n_photons))


def _sums(length: int, total_sum: int) -> Iterable[list[int]]:
    if length == 1:
        yield [
            total_sum,
        ]
    else:
        for value in range(total_sum + 1):
            for permutation in _sums(length - 1, total_sum - value):
                yield [*permutation, value]


def annotated_state_to_string(state: list[list[int]]) -> str:
    """Converts the provided annotated state to a string with ket notation."""
    string = "|"
    for s in state:
        if len(s) > 0:
            string += str(len(s)) + ":("
            for label in s:
                string += str(label) + ","
            string = string[:-1] + "),"
        else:
            string += "0,"
    return string[:-1] + ">"
